# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original authors: Albert Chen, Reza Hosseini
"""Helper functions for
`~greykite.algo.forecast.silverkite.forecast_simple_silverkite.py.`
"""

import warnings

import pandas as pd

from greykite.common import constants as cst
from greykite.common.constants import EVENT_DF_DATE_COL
from greykite.common.constants import EVENT_DF_LABEL_COL
from greykite.common.constants import EVENT_INDICATOR
from greykite.common.features.timeseries_features import add_event_window_multi
from greykite.common.features.timeseries_features import get_fourier_col_name
from greykite.common.features.timeseries_features import get_holidays


def cols_interact(
        static_col,
        fs_name,
        fs_order,
        fs_seas_name=None):
    """Returns all interactions between static_col and fourier series up to specified order

    :param static_col:
        column to interact with fourier series. can be an arbitrary patsy model term
        e.g. "ct1", "C(woy)", "is_weekend:Q('events_Christmas Day')"
    :param fs_name:
        column the fourier series is generated from, same as col_name in fourier_series_fcn
    :param fs_order: int
        generate interactions up to this order. must be <= order in fourier_series_fcn
    :param fs_seas_name: str
        same as seas_name in fourier_series_fcn
    :return: list[str]
        interaction terms to include in patsy model formula
    """
    interaction_columns = [None] * fs_order * 2
    for i in range(fs_order):
        k = i + 1
        sin_col_name = get_fourier_col_name(
            k,
            fs_name,
            function_name="sin",
            seas_name=fs_seas_name)
        cos_col_name = get_fourier_col_name(
            k,
            fs_name,
            function_name="cos",
            seas_name=fs_seas_name)
        interaction_columns[2*i] = f"{static_col}:{sin_col_name}"
        interaction_columns[2*i + 1] = f"{static_col}:{cos_col_name}"
    return interaction_columns


def dedup_holiday_dict(holidays_dict):
    """Removes duplicates from get_holidays output

    :param holidays_dict: dict(str, pd.DataFrame(EVENT_DF_DATE_COL, EVENT_DF_LABEL_COL))
        dictionary from get_holidays
    :return:
        concatenates rows of all DataFrames in holiday_df
        drops duplicate holiday names
    """
    result = pd.DataFrame()
    for country, country_holiday_df in holidays_dict.items():
        result = pd.concat([result, country_holiday_df], axis=0)
    result.drop_duplicates(inplace=True)
    return result


def split_events_into_dictionaries(
        events_df,
        events,
        date_col=EVENT_DF_DATE_COL,
        name_col=EVENT_DF_LABEL_COL,
        default_category="Other"):
    """Splits pd.Dataframe(date, holiday) into separate dataframes, one per event

    Can be used to create the `daily_event_df_dict` parameter for `forecast_silverkite`.
    Each event specified in `events` gets its own effect in the model.
        Other events are grouped together and modeled with the same effect

    :param events_df: pd.DataFrame with date_col and name_col columns
        contains events
    :param events: list(str)
        names of events in events_df.name_col.unique() to split into separate
        dataframes
    :param date_col: str, default "date"
        column in event_df containing the date
    :param name_col: str, default EVENT_DF_LABEL_COL
        column in event_df containing the event name
    :param default_category: str
        name of default event

    :return: dict(label: pd.Dataframe(date_col, name_col))
        with keys = events + [default_category]
            name_col column has a constant value = EVENT_INDICATOR
    """
    result = {}
    # separates rows corresponding to each event into their own dataframe
    for event_name in events:
        event_df = events_df[events_df[name_col] == event_name].copy()
        if event_df.shape[0] > 0:
            event_df[name_col] = EVENT_INDICATOR  # All dates in this df are for the event
            event_key = event_name.replace("'", "")  # ensures patsy module can parse column name in formula
            result[event_key] = event_df.drop_duplicates().reset_index(drop=True)
        else:
            warnings.warn(
                f"Requested holiday '{event_name}' does not occur in the provided countries")

    # groups other events into the same bucket
    other_df = events_df[~events_df[name_col].isin(events)].copy()
    if other_df.shape[0] > 0:
        other_df[name_col] = EVENT_INDICATOR
        default_category = default_category.replace("'", "")
        result[default_category] = other_df.drop_duplicates().reset_index(drop=True)

    # there must be no duplicated dates in each DataFrame
    for k, df in result.items():
        assert not any(df[date_col].duplicated())

    return result


def generate_holiday_events(
        countries,
        holidays_to_model_separately,
        year_start,
        year_end,
        pre_num,
        post_num,
        pre_post_num_dict=None,
        default_category="Other"):
    """Returns holidays within the countries between ``year_start`` and ``year_end``.
    Creates a separate key, value for each item in ``holidays_to_model_separately``.
    The rest are grouped together.

    Useful when multiple countries share the same holiday (e.g. New Year's Day),
    to model a single effect for that holiday.

    Parameters
    ----------
    countries : `list` [`str`]
        Countries of interest.
    holidays_to_model_separately : `list` [`str`]
        Holidays to model.
    year_start: `int`
        Start year for holidays.
    year_end: `int`
        Ending year for holidays.
    pre_num: `int`
        Days to model a separate effect prior to each holiday
    post_num: `int`
        Days to model a separate effect after each holiday.
    pre_post_num_dict : `dict` [`str`, (`int`, `int`)] or None, default None
        Overrides ``pre_num`` and ``post_num`` for each holiday in
        ``holidays_to_model_separately``.
        For example, if ``holidays_to_model_separately`` contains "Thanksgiving" and "Labor Day",
        this parameter can be set to ``{"Thanksgiving": [1, 3], "Labor Day": [1, 2]}``,
        denoting that the "Thanksgiving" ``pre_num`` is 1 and ``post_num`` is 3, and "Labor Day"
        ``pre_num`` is 1 and ``post_num`` is 2.
        Holidays not specified use the default given by ``pre_num`` and ``post_num``.
    default_category: `str`
        Default category name, for holidays in countries not included
        in ``holidays_to_model_separately``.

    Returns
    -------
    daily_event_df_dict : `dict` [`str`, `pandas.DataFrame` (EVENT_DF_DATE_COL, EVENT_DF_LABEL_COL)]
        suitable for use as ``daily_event_df_dict`` parameter in ``forecast_silverkite``
    """
    # retrieves separate DataFrame for each country, with list of holidays
    holidays_dict = get_holidays(
        countries,
        year_start=year_start,
        year_end=year_end)
    if len(holidays_dict) == 0:  # requested holidays are not found the countries
        daily_event_df_dict = None
    else:
        # merges country DataFrames, removes duplicate holidays
        holiday_df = dedup_holiday_dict(holidays_dict)
        # creates separate DataFrame for each holiday
        daily_event_df_dict = split_events_into_dictionaries(
            holiday_df,
            holidays_to_model_separately,
            default_category=default_category)

        # Removes "'" from keys in `pre_post_num_dict` because they are
        # removed from holiday names by ``split_events_into_dictionaries``.
        if pre_post_num_dict:
            # ``.copy()`` is used below to avoid altering the dictionary keys within iteration on same keys
            keys = pre_post_num_dict.copy().keys()
            for key in keys:
                new_key = key.replace("'", "")
                if key not in daily_event_df_dict:
                    pre_post_num_dict[new_key] = pre_post_num_dict.pop(key)
                if new_key not in daily_event_df_dict:
                    warnings.warn(
                        f"Requested holiday '{new_key}' is not valid. Valid holidays are: "
                        f"{list(daily_event_df_dict.keys())}", UserWarning)

        shifted_event_dict = add_event_window_multi(
            event_df_dict=daily_event_df_dict,
            time_col=EVENT_DF_DATE_COL,
            label_col=EVENT_DF_LABEL_COL,
            time_delta="1D",
            pre_num=pre_num,
            post_num=post_num,
            pre_post_num_dict=pre_post_num_dict)

        daily_event_df_dict.update(shifted_event_dict)
    return daily_event_df_dict


def patsy_categorical_term(
        term,
        levels=None,
        coding=None,
        quote=True):
    """Returns categorical term for patsy.
        Optionally specify levels, coding, and quote the term
    :param term: str
        name of the categorical variable
    :param levels: list(str) or None
        levels for the categorical variable
    :param coding: str
        A valid coding. E.g. Treatment, Sum, Diff, Poly
        https://patsy.readthedocs.io/en/latest/API-reference.html#handling-categorical-data
    :param quote: bool
        whether to quote the term. Useful if there is a space or "." in the term
    :return: str
        categorical factor for patsy model formula
    """
    if quote:
        term = f"Q('{term}')"
    # constructs the string for the patsy model term
    string = f"C({term}"
    if coding is not None:
        string += f", {coding}"
    if levels is not None:
        string += f", levels={levels}"
    string += ")"
    return string


def get_event_pred_cols(daily_event_df_dict):
    """Generates the names of internal predictor columns from
    the event dictionary passed to
    `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`.
    These can be passed via the ``extra_pred_cols`` parameter to model event effects.

    .. note::

        The returned strings are patsy model formula terms.
        Each provides full set of levels so that prediction works even if a
        level is not found in the training set.

        If a level does not appear in the training set, its coefficient may
        be unbounded in the "linear" fit_algorithm. A method with regularization
        avoids this issue (e.g. "ridge", "elastic_net").

    Parameters
    ----------
    daily_event_df_dict : `dict` or None, optional, default None
        A dictionary of data frames, each representing events data for the corresponding key.
        See `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`.

    Returns
    -------
    event_pred_cols : `list` [`str`]
        List of patsy model formula terms, one for each
        key of ``daily_event_df_dict``.
    """
    event_pred_cols = []
    if daily_event_df_dict is not None:
        for key in sorted(daily_event_df_dict.keys()):
            # `add_daily_events` creates a column with this name.
            term = f"{cst.EVENT_PREFIX}_{key}"
            # Its values are set to the event df label column. Dates that do not correspond
            # to the event are set to `cst.EVENT_DEFAULT`.
            event_levels = [cst.EVENT_DEFAULT]  # reference level for non-event days
            event_levels += list(daily_event_df_dict[key][cst.EVENT_DF_LABEL_COL].unique())  # this event's levels
            event_pred_cols += [patsy_categorical_term(term=term, levels=event_levels)]
    return event_pred_cols



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Kaixu Yang
"""Provides summaries of sklearn and statsmodels
regression models.
"""

from greykite.algo.common.col_name_utils import filter_coef_summary
from greykite.algo.common.model_summary_utils import format_summary_df
from greykite.algo.common.model_summary_utils import get_info_dict_lm
from greykite.algo.common.model_summary_utils import get_info_dict_tree
from greykite.algo.common.model_summary_utils import print_summary
from greykite.algo.common.model_summary_utils import process_intercept


class ModelSummary:
    """A class to store regression model summary statistics.

    The class can be printed to get a well formatted model summary.

    Attributes
    ----------
    x : `numpy.array`
        The design matrix.
    beta : `numpy.array`
        The estimated coefficients.
    y : `numpy.array`
        The response.
    pred_cols : `list` [ `str` ]
        List of predictor names.
    pred_category : `dict`
        Predictor category, returned by
        `~greykite.algo.common.col_name_utils.create_pred_category`.
    fit_algorithm : `str`
        The name of algorithm to fit the regression.
    ml_model : `class`
        The trained machine learning model class.
    max_colwidth : `int`
        The maximum length for predictors to be shown in their original name.
        If the maximum length of predictors exceeds this parameter, all
        predictors name will be suppressed and only indices are shown.
    info_dict : `dict`
        The model summary dictionary, output of
        `~greykite.algo.common.model_summary.ModelSummary._get_summary`
    """

    def __init__(
            self,
            x,
            y,
            pred_cols,
            pred_category,
            fit_algorithm,
            ml_model,  # needs to support cloning if ml_model implements a lasso method
            max_colwidth=20):
        # process x, beta and pred_cols to include the intercept term
        beta = getattr(ml_model, "coef_", None)
        intercept = getattr(ml_model, "intercept_", None)
        if beta is not None:
            x, beta, pred_cols = process_intercept(x, beta, intercept, pred_cols)
        self.x = x
        self.beta = beta
        self.y = y
        self.pred_cols = pred_cols
        self.pred_category = pred_category
        self.fit_algorithm = fit_algorithm
        self.ml_model = ml_model
        self.max_colwidth = max_colwidth
        self.info_dict = self._get_summary()

    def __str__(self):
        """print method.
        """
        return print_summary(self.info_dict, self.max_colwidth)

    def __repr__(self):
        """print method
        """
        return print_summary(self.info_dict, self.max_colwidth)

    def _get_summary(self):
        """Gets the model summary from input.
        This function is called during initialization.

        Returns
        -------
        info_dict : `dict`
            Includes direct and derived metrics about the trained model. For detailed keys, refer to
            `~greykite.algo.common.model_summary_utils.get_info_dict_lm`
            or
            `~greykite.algo.common.model_summary_utils.get_info_dict_tree`.
        """
        if self.fit_algorithm in ["linear", "ridge", "lasso", "lars", "lasso_lars",
                                  "sgd", "elastic_net", "statsmodels_ols",
                                  "statsmodels_wls", "statsmodels_gls", "statsmodels_glm"]:
            info_dict = get_info_dict_lm(
                x=self.x,
                y=self.y,
                beta=self.beta,
                ml_model=self.ml_model,
                fit_algorithm=self.fit_algorithm,
                pred_cols=self.pred_cols)
        elif self.fit_algorithm in ["rf", "gradient_boosting"]:
            info_dict = get_info_dict_tree(
                x=self.x,
                y=self.y,
                ml_model=self.ml_model,
                fit_algorithm=self.fit_algorithm,
                pred_cols=self.pred_cols)
        else:
            raise NotImplementedError(f"{self.fit_algorithm} is not recognized, "
                                      f"summary is not implemented.")
        return info_dict

    def get_coef_summary(
            self,
            is_intercept=None,
            is_time_feature=None,
            is_event=None,
            is_trend=None,
            is_seasonality=None,
            is_lag=None,
            is_regressor=None,
            is_interaction=None,
            return_df=False):
        """Gets the coefficient summary filtered by conditions.

        Parameters
        ----------
        is_intercept : `bool` or `None`, default `None`
            Intercept or not.
        is_time_feature : `bool` or `None`, default `None`
            Time features or not.
            Time features belong to `~greykite.common.constants.TIME_FEATURES`.
        is_event : `bool` or `None`, default `None`
            Event features or not.
            Event features have `~greykite.common.constants.EVENT_PREFIX`.
        is_trend : `bool` or `None`, default `None`
            Trend features or not.
            Trend features have `~greykite.common.constants.CHANGEPOINT_COL_PREFIX` or "cp\\d".
        is_seasonality : `bool` or `None`, default `None`
            Seasonality feature or not.
            Seasonality features have `~greykite.common.constants.SEASONALITY_REGEX`.
        is_lag : `bool` or `None`, default `None`
            Lagged features or not.
            Lagged features have "lag".
        is_regressor : 0 or 1
            Extra features provided by users.
            They are provided through ``extra_pred_cols`` in the fit function.
        is_interaction : `bool` or `None`, default `None`
            Interaction feature or not.
            Interaction features have ":".
        return_df : `bool`, default `False`
           If True, the filtered coefficient summary df is also returned.
            Otherwise, the filtered coefficient summary df is printed only.

        Returns
        -------
        filtered_coef_summary : `pandas.DataFrame` or `None`
            If ``return_df`` is set to True, returns the filtered coefficient summary
            df filtered by the given conditions.
        """
        filtered_coef_summary = filter_coef_summary(
            coef_summary=self.info_dict["coef_summary_df"],
            pred_category=self.pred_category,
            is_intercept=is_intercept,
            is_time_feature=is_time_feature,
            is_event=is_event,
            is_trend=is_trend,
            is_seasonality=is_seasonality,
            is_lag=is_lag,
            is_regressor=is_regressor,
            is_interaction=is_interaction)
        # excludes column categories
        cols = [col for col in filtered_coef_summary.columns if "is_" not in col]
        print(format_summary_df(filtered_coef_summary[cols]).to_string(index=False))
        if return_df:
            return filtered_coef_summary



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Sayan Patra
"""Class to load datasets."""
import datetime
import os
from pathlib import Path

import pandas as pd

from greykite.common.constants import TIME_COL


class DataLoader:
    """Returns datasets included in the library in `pandas.DataFrame` format.

    Attributes
    ----------
    available_datasets : `list` [`str`]
        The names of the available datasets.
    """
    def __init__(self):
        self.available_datasets = self.get_data_inventory()

    @staticmethod
    def get_data_home(data_dir=None, data_sub_dir=None):
        """Returns the folder path ``data_dir/data_sub_dir``.
        If ``data_dir`` is None returns the internal data directory.
        By default the Greykite data dir is set to a folder named 'data' in the project source code.
        Alternatively, it can be set programmatically by giving an explicit folder path.

        Parameters
        ----------
        data_dir : `str` or None, default None
            The path to the input data directory.
            If None, it is set to '{project_trunk}/data'.
        data_sub_dir : `str` or None, default None
            The name of the input data sub directory.
            Updates path by appending to the ``data_dir`` at the end.
            If None, ``data_dir`` path is unchanged.

        Returns
        -------
        data_home : `str`
            Path to the data folder.
        """
        if data_dir is None:
            data_dir = Path(__file__).parents[1].joinpath("data")

        if data_sub_dir is None:
            data_home = os.path.abspath(data_dir)
        else:
            data_home = os.path.abspath(os.path.join(data_dir, data_sub_dir))

        if not os.path.exists(data_home):
            raise ValueError(f"Requested data directory '{data_home}' does not exist.")

        return data_home

    @staticmethod
    def get_data_names(data_path):
        """Returns the names of the ``.csv`` and ``.csv.xz`` files in ``data_path``.

        Parameters
        ----------
        data_path : `str`
            Path to the data folder.

        Returns
        -------
        file_names : `list` [`str`]
            The names of the ``.csv`` and ``.csv.xz`` files in ``data_path``.
        """
        file_names = os.listdir(data_path)
        file_names = [file_name.split(".csv")[0] for file_name in file_names
                      if file_name.endswith((".csv", ".csv.xz"))]

        return file_names

    def get_data_inventory(self):
        """Returns the names of the available internal datasets.

        Returns
        -------
        file_names : `list` [`str`]
            The names of the available internal datasets.
        """
        file_names = []
        for freq in ["minute", "hourly", "daily", "monthly"]:
            data_path = self.get_data_home(data_dir=None, data_sub_dir=freq)
            file_names.extend(self.get_data_names(data_path=data_path))

        return file_names

    def get_df(self, data_path, data_name):
        """Returns a ``pandas.DataFrame`` containing the dataset from ``data_path/data_name``.
        The input data must be in ``.csv`` or ``.csv.xz`` format.
        Raises a ValueError if the the specified input file is not found.

        Parameters
        ----------
        data_path : `str`
            Path to the data folder.
        data_name : `str`
            Name of the csv file to be loaded from. For example 'peyton_manning'.

        Returns
        -------
        df : `pandas.DataFrame`
            Input dataset.
        """
        file_path = os.path.join(data_path, f"{data_name}.csv")
        if not os.path.exists(file_path):
            file_names = self.get_data_names(data_path)
            raise ValueError(f"Given file path '{file_path}' is not found. Available datasets in "
                             f"data directory '{data_path}' are {file_names}.")
        df = pd.read_csv(file_path, sep=",")
        return df

    def load_peyton_manning(self):
        """Loads the Daily Peyton Manning dataset.

        This dataset contains log daily page views for the Wikipedia page for Peyton Manning.
        One of the primary datasets used for demonstrations by Facebook ``Prophet`` algorithm.
        Source: https://github.com/facebook/prophet/blob/master/examples/example_wp_log_peyton_manning.csv

        Below is the dataset attribute information:

            ts : date of the page view
            y : log of the number of page views

        Returns
        -------
        df : `pandas.DataFrame` object with Peyton Manning data.
            Has the following columns:

                "ts" : date of the page view.
                "y" : log of the number of page views.
        """
        data_path = self.get_data_home(data_dir=None, data_sub_dir="daily")
        df = self.get_df(data_path=data_path, data_name="daily_peyton_manning")
        return df

    def load_parking(self, system_code_number=None):
        """Loads the Hourly Parking dataset.
        This dataset contains occupancy rates (8:00 to 16:30) from 2016/10/04 to 2016/12/19
        from car parks in Birmingham that are operated by NCP from Birmingham City Council.
        Source: https://archive.ics.uci.edu/ml/datasets/Parking+Birmingham
        UK Open Government Licence (OGL)

        Below is the dataset attribute information:

            SystemCodeNumber: car park ID
            Capacity: car park capacity
            Occupancy: car park occupancy rate
            LastUpdated: date and time of the measure

        Parameters
        ----------
        system_code_number : `str` or None, default None
            If None, occupancy rate is averaged across all the ``SystemCodeNumber``.
            Else only the occupancy rate of the given ``system_code_number`` is returned.

        Returns
        -------
        df : `pandas.DataFrame` object with Parking data.
            Has the following columns:

                "LastUpdated" : time, rounded to the nearest half hour.
                "Capacity" : car park capacity
                "Occupancy" : car park occupancy rate
                "OccupancyRatio" : ``Occupancy`` divided by ``Capacity``.
        """
        data_path = self.get_data_home(data_dir=None, data_sub_dir="hourly")
        df = self.get_df(data_path=data_path, data_name="hourly_parking")

        # Rounds time column to nearest half hour point
        df["LastUpdated"] = pd.to_datetime(df["LastUpdated"]).dt.round("30min")
        df["OccupancyRatio"] = df["Occupancy"] / df["Capacity"]
        if system_code_number is None:
            df = df.groupby("LastUpdated", as_index=False).mean()
        else:
            df = df[df["SystemCodeNumber"] == system_code_number].reset_index(drop=True)
        return df

    def load_bikesharing(self):
        """Loads the Hourly Bike Sharing Count dataset.

        This dataset contains aggregated hourly count of the number of rented bikes.
        The data also includes weather data: Maximum Daily temperature (tmax);
        Minimum Daily Temperature (tmin); Precipitation (pn)
        The raw bike-sharing data is provided by Capital Bikeshares.
        Source: https://www.capitalbikeshare.com/system-data
        The raw weather data (Baltimore-Washington INTL Airport)
        https://www.ncdc.noaa.gov/data-access/land-based-station-data

        Below is the dataset attribute information:

            ts : hour and date
            count : number of shared bikes
            tmin : minimum daily temperature
            tmax : maximum daily temperature
            pn : precipitation

        Returns
        -------
        df : `pandas.DataFrame` with bikesharing data.
            Has the following columns:

                "date" : day of year
                "ts" : hourly timestamp
                "count" : number of rented bikes across Washington DC.
                "tmin" : minimum daily temperature
                "tmax" : maximum daily temperature
                "pn" : precipitation
        """
        data_path = self.get_data_home(data_dir=None, data_sub_dir="hourly")
        df = self.get_df(data_path=data_path, data_name="hourly_bikesharing")
        return df

    def load_beijing_pm(self):
        """Loads the Beijing Particulate Matter (PM2.5) dataset.
        https://archive.ics.uci.edu/ml/datasets/Beijing+PM2.5+Data

        This hourly data set contains the PM2.5 data of US Embassy in Beijing. Meanwhile, meteorological data
        from Beijing Capital International Airport are also included.

        The dataset's time period is between Jan 1st, 2010 to Dec 31st, 2014. Missing data are denoted as NA.

        Below is the dataset attribute information:

            No : row number
            year : year of data in this row
            month : month of data in this row
            day : day of data in this row
            hour : hour of data in this row
            pm2.5: PM2.5 concentration (ug/m^3)
            DEWP : dew point (celsius)
            TEMP : temperature (celsius)
            PRES : pressure (hPa)
            cbwd : combined wind direction
            Iws : cumulated wind speed (m/s)
            Is : cumulated hours of snow
            Ir : cumulated hours of rain

        Returns
        -------
        df : `pandas.DataFrame` with Beijing PM2.5 data.
            Has the following columns:

                TIME_COL : hourly timestamp
                "year" : year of data in this row
                "month" : month of data in this row
                "day" : day of data in this row
                "hour" : hour of data in this row
                "pm" : PM2.5 concentration (ug/m^3)
                "dewp" : dew point (celsius)
                "temp" : temperature (celsius)
                "pres" : pressure (hPa)
                "cbwd" : combined wind direction
                "iws" : cumulated wind speed (m/s)
                "is" : cumulated hours of snow
                "ir" : cumulated hours of rain
        """
        data_path = self.get_data_home(data_dir=None, data_sub_dir="hourly")
        df = self.get_df(data_path=data_path, data_name="hourly_beijing_pm")

        df.drop("No", axis=1, inplace=True)
        df.columns = map(str.lower, df.columns)
        df.rename(columns={"pm2.5": "pm"}, inplace=True)

        hour_string = df["year"].astype(str) + "-" + df["month"].astype(str) + "-" + df["day"].astype(str) + "-" + df["hour"].astype(str)
        hour_datetime = hour_string.apply(lambda x: datetime.datetime.strptime(x, "%Y-%m-%d-%H"))
        df.insert(0, TIME_COL, hour_datetime)
        return df

    def load_data(self, data_name, **kwargs):
        """Loads dataset by name from the internal data library.

        Parameters
        ----------
        data_name : `str`
            Dataset to load from the internal data library.

        Returns
        -------
        df : ``UnivariateTimeSeries`` object with ``data_name``.
        """
        if data_name == "daily_peyton_manning":
            df = self.load_peyton_manning()
        elif data_name == "hourly_parking":
            df = self.load_parking(**kwargs)
        elif data_name == "hourly_bikesharing":
            df = self.load_bikesharing()
        elif data_name == "hourly_beijing_pm":
            df = self.load_beijing_pm()
        else:
            data_inventory = self.get_data_inventory()
            raise ValueError(f"Input data name '{data_name}' is not recognized. "
                             f"Must be one of {data_inventory}.")

        return df



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Rachit Kumar, Albert Chen
"""Template for `greykite.sklearn.estimator.prophet_estimator`.
Takes input data and forecast config,
and returns parameters to call
:func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.
"""

import dataclasses
import functools
import warnings
from typing import Dict
from typing import Optional

import pandas as pd
from fbprophet.make_holidays import make_holidays_df

from greykite.common.python_utils import dictionaries_values_to_lists
from greykite.common.python_utils import update_dictionaries
from greykite.common.python_utils import update_dictionary
from greykite.framework.templates.autogen.forecast_config import ForecastConfig
from greykite.framework.templates.autogen.forecast_config import ModelComponentsParam
from greykite.framework.templates.base_template import BaseTemplate
from greykite.sklearn.estimator.base_forecast_estimator import BaseForecastEstimator
from greykite.sklearn.estimator.prophet_estimator import ProphetEstimator


class ProphetTemplate(BaseTemplate):
    """A template for :class:`~greykite.sklearn.estimator.prophet_estimator.ProphetEstimator`.

    Takes input data and optional configuration parameters
    to customize the model. Returns a set of parameters to call
    :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.

    Notes
    -----
    The attributes of a `~greykite.framework.templates.autogen.forecast_config.ForecastConfig` for
    :class:`~greykite.sklearn.estimator.prophet_estimator.ProphetEstimator` are:

        computation_param: `ComputationParam` or None, default None
            How to compute the result. See
            :class:`~greykite.framework.templates.autogen.forecast_config.ComputationParam`.
        coverage: `float` or None, default None
            Intended coverage of the prediction bands (0.0 to 1.0)
            If None, the upper/lower predictions are not returned
            Same as coverage in ``forecast_pipeline``
        evaluation_metric_param: `EvaluationMetricParam` or None, default None
            What metrics to evaluate. See
            :class:`~greykite.framework.templates.autogen.forecast_config.EvaluationMetricParam`.
        evaluation_period_param: `EvaluationPeriodParam` or None, default None
            How to split data for evaluation. See
            :class:`~greykite.framework.templates.autogen.forecast_config.EvaluationPeriodParam`.
        forecast_horizon: `int` or None, default None
            Number of periods to forecast into the future. Must be > 0
            If None, default is determined from input data frequency
            Same as forecast_horizon in `forecast_pipeline`
        metadata_param: `MetadataParam` or None, default None
            Information about the input data. See
            :class:`~greykite.framework.templates.autogen.forecast_config.MetadataParam`.
        model_components_param: `ModelComponentsParam` or None, default None
            Parameters to tune the model. See
            :class:`~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam`.
            The fields are dictionaries with the following items.

            seasonality: `dict` [`str`, `any`] or None
                Seasonality config dictionary, with the following optional keys.

                ``"seasonality_mode"``: `str` or None or list of such values for grid search
                    Can be 'additive' (default) or 'multiplicative'.
                ``"seasonality_prior_scale"``: `float` or None or list of such values for grid search
                    Parameter modulating the strength of the seasonality model.
                    Larger values allow the model to fit larger seasonal fluctuations, smaller values dampen the seasonality.
                    Specify for individual seasonalities using add_seasonality_dict.
                ``"yearly_seasonality"``: `str` or `bool` or `int` or list of such values for grid search, default 'auto'
                    Determines the yearly seasonality
                    Can be 'auto', True, False, or a number of Fourier terms to generate.
                ``"weekly_seasonality"``: `str` or `bool` or `int` or list of such values for grid search, default 'auto'
                    Determines the weekly seasonality
                    Can be 'auto', True, False, or a number of Fourier terms to generate.
                ``"daily_seasonality"``: `str` or `bool` or `int` or list of such values for grid search, default 'auto'
                    Determines the daily seasonality
                    Can be 'auto', True, False, or a number of Fourier terms to generate.
                ``"add_seasonality_dict"``: `dict` or None or list of such values for grid search
                    dict of custom seasonality parameters to be added to the model, default=None
                    Key is the seasonality component name e.g. 'monthly'; parameters are specified via dict.
                    See :class:`~greykite.sklearn.estimator.prophet_estimator` for details.

            growth: `dict` [`str`, `any`] or None
                Specifies the growth parameter configuration.
                Dictionary with the following optional key:

                ``"growth_term"``: `str` or None or list of such values for grid search
                    How to model the growth. Valid options are "linear" and "logistic"
                    Specify a linear or logistic trend, these terms have their origin at the train start date.

            events: `dict` [`str`, `any`] or None
                Holiday/events configuration dictionary with the following optional keys:

                ``"holiday_lookup_countries"``: `list` [`str`] or "auto" or None
                    Which countries' holidays to include. Must contain all the holidays you intend to model.
                    If "auto", uses default list of countries with large contribution to Internet traffic.
                    If None or an empty list, no holidays are modeled.
                ``"holidays_prior_scale"``: `float` or None or list of such values for grid search, default 10.0
                    Modulates the strength of the holiday effect.
                ``"holiday_pre_num_days"``: `list` [`int`] or None, default 2
                    Model holiday effects for holiday_pre_num_days days before the holiday.
                    Grid search is not supported. Must be a list with one element or None.
                ``"holiday_post_num_days"``: `list` [`int`] or None, default 2
                    Model holiday effects for holiday_post_num_days days after the holiday
                    Grid search is not supported. Must be a list with one element or None.

            changepoints: `dict` [`str`, `any`] or None
                Specifies the changepoint configuration. Dictionary with the following
                optional keys:

                ``"changepoint_prior_scale"`` : `float` or None or list of such values for grid search, default 0.05
                    Parameter modulating the flexibility of the automatic changepoint selection. Large values will allow many
                    changepoints, small values will allow few changepoints.
                ``"changepoints"`` : `list` [`datetime.datetime`] or None or list of such values for grid search, default None
                    List of dates at which to include potential changepoints. If not specified,
                    potential changepoints are selected automatically.
                ``"n_changepoints"`` : `int` or None or list of such values for grid search, default 25
                    Number of potential changepoints to include. Not used if input `changepoints` is supplied.
                    If `changepoints` is not supplied, then n_changepoints potential changepoints are selected uniformly from
                    the first `changepoint_range` proportion of the history.
                ``"changepoint_range"`` : `float` or None or list of such values for grid search, default 0.8
                    Proportion of history in which trend changepoints will be estimated. Permitted values: (0,1]
                    Not used if input `changepoints` is supplied.

            regressors: `dict` [`str`, `any`] or None
                Specifies the regressors to include in the model (e.g. macro-economic factors).
                Dictionary with the following optional keys:

                ``"add_regressor_dict"`` : `dict` or None or list of such values for grid search, default None
                    Dictionary of extra regressors to be modeled.
                    See `~greykite.sklearn.estimator.prophet_estimator.ProphetEstimator` for details.

            uncertainty: `dict` [`str`, `any`] or None
                Specifies the uncertainty configuration. A dictionary with the following optional keys:

                ``"mcmc_samples"`` : `int` or None or list of such values for grid search, default 0
                    if greater than 0, will do full Bayesian inference with the specified number of MCMC samples.
                    If 0, will do MAP estimation.
                ``"uncertainty_samples"`` : `int` or None or list of such values for grid search, default 1000
                    Number of simulated draws used to estimate
                    uncertainty intervals. Setting this value to 0 or False will disable
                    uncertainty estimation and speed up the calculation.

            hyperparameter_override: `dict` [`str`, `any`] or None or `list` [`dict` [`str`, `any`] or None]
                After the above model components are used to create a hyperparameter grid, the result is
                updated by this dictionary, to create new keys or override existing ones.
                Allows for complete customization of the grid search.

                Keys should have format ``{named_step}__{parameter_name}`` for the named steps of the
                `sklearn.pipeline.Pipeline` returned by this function. See `sklearn.pipeline.Pipeline`.

                For example::

                    hyperparameter_override={
                        "estimator__yearly_seasonality": [True, False],
                        "estimator__seasonality_prior_scale": [5.0, 15.0],
                        "input__response__null__impute_algorithm": "ts_interpolate",
                        "input__response__null__impute_params": {"orders": [7, 14]},
                        "input__regressors_numeric__normalize__normalize_algorithm": "RobustScaler",
                    }

                If a list of dictionaries, grid search will be done for each dictionary in the list.
                Each dictionary in the list override the defaults. This enables grid search
                over specific combinations of parameters to reduce the search space.

                    * For example, the first dictionary could define combinations of parameters for a
                      "complex" model, and the second dictionary could define combinations of parameters
                      for a "simple" model, to prevent mixed combinations of simple and complex.
                    * Or the first dictionary could grid search over fit algorithm, and the second dictionary
                      could use a single fit algorithm and grid search over seasonality.

                The result is passed as the ``param_distributions`` parameter
                to `sklearn.model_selection.RandomizedSearchCV`.
            autoregression: `dict` [`str`, `any`] or None
                Ignored. Prophet template does not support autoregression.
            custom: `dict` [`str`, `any`] or None
                Ignored. There are no custom options.
        model_template: `str`
            This class only accepts "PROPHET".
    """
    DEFAULT_MODEL_TEMPLATE = "PROPHET"
    """The default model template. See `~greykite.framework.templates.model_templates.ModelTemplateEnum`.
    Uses a string to avoid circular imports.
    Overrides the value from `~greykite.framework.templates.forecast_config_defaults.ForecastConfigDefaults`.
    """
    HOLIDAY_LOOKUP_COUNTRIES_AUTO = (
        "UnitedStates", "UnitedKingdom", "India", "France", "China")
    """Default holiday countries to use if countries='auto'"""
    def __init__(
            self,
            estimator: BaseForecastEstimator = ProphetEstimator()):
        super().__init__(estimator=estimator)

    @property
    def allow_model_template_list(self):
        """ProphetTemplate does not allow `config.model_template` to be a list."""
        return False

    @property
    def allow_model_components_param_list(self):
        """ProphetTemplate does not allow `config.model_components_param` to be a list."""
        return False

    def get_prophet_holidays(
            self,
            year_list,
            countries="auto",
            lower_window=-2,
            upper_window=2):
        """Generates holidays for Prophet model.

        Parameters
        ----------
        year_list : `list` [`int`]
            List of years for selecting the holidays across given countries.
        countries : `list` [`str`] or "auto" or None, default "auto"
            Countries for selecting holidays.

            * If "auto", uses Top Countries for internet traffic.
            * If a list, a list of country names.
            * If None, the function returns None.

        lower_window : `int` or None, default -2
            Negative integer. Model holiday effects for given number of days before the holiday.
        upper_window : `int` or None, default 2
            Positive integer. Model holiday effects for given number of days after the holiday.

        Returns
        -------
        holidays : `pandas.DataFrame`
            holidays dataframe to pass to Prophet's `holidays` argument.

        See Also
        --------
        `~greykite.common.features.timeseries_features.get_available_holiday_lookup_countries`
        to list available countries for modeling.

        `~greykite.sklearn.estimator.prophet_estimator.ProphetEstimator`.
        """
        holidays = None
        if countries is None:
            countries = []
        elif countries == "auto":
            # countries with large contribution to internet traffic
            countries = self.HOLIDAY_LOOKUP_COUNTRIES_AUTO
        elif not isinstance(countries, (list, tuple)):
            raise ValueError(f"`countries` should be a list, found {countries}")

        # Suppresses the warnings such as "We only support Diwali and Holi holidays from 2010 to 2025"
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            for country in countries:
                country_holidays = make_holidays_df(year_list=year_list, country=country)
                country_holidays["lower_window"] = lower_window
                country_holidays["upper_window"] = upper_window
                if holidays is not None:
                    holidays = pd.concat([holidays, country_holidays], axis=0)
                else:
                    holidays = country_holidays
        if holidays is not None:
            holidays.drop_duplicates(inplace=True)
        return holidays

    def get_regressor_cols(self):
        """Returns regressor column names.

        Implements the method in `~greykite.framework.templates.base_template.BaseTemplate`.

        Returns
        -------
        regressor_cols : `list` [`str`] or None
            The names of regressor columns used in any hyperparameter set
            requested by ``model_components``.
            None if there are no regressors.
        """
        # ``add_regressor_dict`` is a list of dictionaries to initialize
        # ProphetEstimator.add_seasonality_dict. Each dictionary's keys are
        # the regressors used in that model
        reg_names = set()
        if self.config.model_components_param.regressors is not None:
            add_regressor_dict = self.config.model_components_param.regressors.get("add_regressor_dict", [None])
            if isinstance(add_regressor_dict, dict):
                add_regressor_dict = [add_regressor_dict]
            for reg_dict in add_regressor_dict:
                # None indicates no regressors
                if reg_dict is not None:
                    reg_names.update(reg_dict.keys())
        regressor_cols = list(reg_names) if reg_names else None
        return regressor_cols

    def apply_prophet_model_components_defaults(
            self,
            model_components=None,
            time_properties=None):
        """Sets default values for ``model_components``.

        Called by ``get_hyperparameter_grid`` after ``time_properties` is defined.
        Requires ``time_properties`` as well as ``model_components``
        so we do not simply override
        `~greykite.framework.templates.forecast_config_defaults.ForecastConfigDefaults.apply_model_components_defaults`.

        Parameters
        ----------
        model_components : :class:`~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam` or None, default None
            Configuration of model growth, seasonality, events, etc.
            See the docstring of this class for details.
        time_properties : `dict` [`str`, `any`] or None, default None
            Time properties dictionary (likely produced by
            `~greykite.common.time_properties_forecast.get_forecast_time_properties`)
            with keys:

                ``"period"`` : `int`
                    Period of each observation (i.e. minimum time between observations, in seconds).
                ``"simple_freq"`` : `SimpleTimeFrequencyEnum`
                    ``SimpleTimeFrequencyEnum`` member corresponding to data frequency.
                ``"num_training_points"`` : `int`
                    Number of observations for training.
                ``"num_training_days"`` : `int`
                    Number of days for training.
                ``"start_year"`` : `int`
                    Start year of the training period.
                ``"end_year"`` : `int`
                    End year of the forecast period.
                ``"origin_for_time_vars"`` : `float`
                    Continuous time representation of the first date in ``df``.

            If None, start_year is set to 2015 and end_year to 2030.

        Returns
        -------
        model_components : :class:`~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam`
            The provided ``model_components`` with default values set
        """
        if model_components is None:
            model_components = ModelComponentsParam()
        else:
            # makes a copy to avoid mutating input
            model_components = dataclasses.replace(model_components)
        if time_properties is None:
            time_properties = {
                "start_year": 2015,
                "end_year": 2030,
            }

        # seasonality
        default_seasonality = {
            "seasonality_mode": ["additive"],
            "seasonality_prior_scale": [10.0],
            "yearly_seasonality": ['auto'],
            "weekly_seasonality": ['auto'],
            "daily_seasonality": ['auto'],
            "add_seasonality_dict": [None]
        }
        # If seasonality params are not provided, uses default params. Otherwise, prefers provided params.
        # `allow_unknown_keys=False` requires `model_components.seasonality` keys to be a subset of
        # `default_seasonality` keys.
        model_components.seasonality = update_dictionary(
            default_dict=default_seasonality,
            overwrite_dict=model_components.seasonality,
            allow_unknown_keys=False)

        # growth
        default_growth = {
            "growth_term": ["linear"]
        }
        model_components.growth = update_dictionary(
            default_dict=default_growth,
            overwrite_dict=model_components.growth,
            allow_unknown_keys=False)

        # events
        default_events = {
            "holiday_lookup_countries": "auto",  # see `get_prophet_holidays` for defaults
            "holiday_pre_num_days": [2],
            "holiday_post_num_days": [2],
            "start_year": time_properties["start_year"],
            "end_year": time_properties["end_year"],
            "holidays_prior_scale": [10.0]
        }
        model_components.events = update_dictionary(
            default_dict=default_events,
            overwrite_dict=model_components.events,
            allow_unknown_keys=False)

        # Creates events dictionary for prophet estimator
        # Expands the range of holiday years by 1 year on each end, to ensure we have coverage of most relevant holidays.
        year_list = list(range(
            model_components.events["start_year"]-1,
            model_components.events["end_year"]+2))
        # Currently we support only one set of holiday_lookup_countries, holiday_pre_num_days and holiday_post_num_days.
        # Shows a warning if user supplies >1 set.
        if len(model_components.events["holiday_pre_num_days"]) > 1:
            warnings.warn(
                f"`events['holiday_pre_num_days']` list has more than 1 element. We currently support only 1 element. "
                f"Using {model_components.events['holiday_pre_num_days'][0]}.")
        if len(model_components.events["holiday_post_num_days"]) > 1:
            warnings.warn(
                f"`events['holiday_post_num_days']` list has more than 1 element. We currently support only 1 element. "
                f"Using {model_components.events['holiday_post_num_days'][0]}.")
        # If events["holiday_lookup_countries"] has multiple options, picks the first option
        if (model_components.events["holiday_lookup_countries"] is not None
                and model_components.events["holiday_lookup_countries"] != "auto"):
            if len(model_components.events["holiday_lookup_countries"]) > 1:
                # There are multiple elements
                if (any(isinstance(x, list) for x in model_components.events["holiday_lookup_countries"])
                        or None in model_components.events["holiday_lookup_countries"]
                        or "auto" in model_components.events["holiday_lookup_countries"]):
                    # Not a flat list of country names
                    warnings.warn(
                        f"`events['holiday_lookup_countries']` contains multiple options. "
                        f"We currently support only 1 option. Using {model_components.events['holiday_lookup_countries'][0]}.")
                    model_components.events["holiday_lookup_countries"] = model_components.events["holiday_lookup_countries"][0]
            elif isinstance(model_components.events["holiday_lookup_countries"][0], (list, tuple)):
                # There's only one element, and it's a list of countries
                model_components.events["holiday_lookup_countries"] = model_components.events["holiday_lookup_countries"][0]

        model_components.events = {
            "holidays_df": self.get_prophet_holidays(
                year_list=year_list,
                countries=model_components.events["holiday_lookup_countries"],
                # holiday effect is modeled from "holiday_pre_num_days" days before
                # to "holiday_post_num_days" days after the holiday
                lower_window=-model_components.events["holiday_pre_num_days"][0],  # Prophet expects a negative value for `lower_window`
                upper_window=model_components.events["holiday_post_num_days"][0]),
            "holidays_prior_scale": model_components.events["holidays_prior_scale"]
        }

        # changepoints_dict
        default_changepoints = {
            "changepoint_prior_scale": [0.05],
            "changepoints": [None],
            "n_changepoints": [25],
            "changepoint_range": [0.8]
        }
        model_components.changepoints = update_dictionary(
            default_dict=default_changepoints,
            overwrite_dict=model_components.changepoints,
            allow_unknown_keys=False)

        # uncertainty
        default_uncertainty = {
            "mcmc_samples": [0],
            "uncertainty_samples": [1000]
        }
        model_components.uncertainty = update_dictionary(
            default_dict=default_uncertainty,
            overwrite_dict=model_components.uncertainty,
            allow_unknown_keys=False)

        # regressors
        default_regressors = {
            "add_regressor_dict": [None]
        }
        model_components.regressors = update_dictionary(
            default_dict=default_regressors,
            overwrite_dict=model_components.regressors,
            allow_unknown_keys=False)

        # there are no custom parameters for Prophet

        # sets to {} if None, for each item if
        # `model_components.hyperparameter_override` is a list of dictionaries
        model_components.hyperparameter_override = update_dictionaries(
            {},
            overwrite_dicts=model_components.hyperparameter_override)

        return model_components

    def get_hyperparameter_grid(self):
        """Returns hyperparameter grid.

        Implements the method in `~greykite.framework.templates.base_template.BaseTemplate`.

        Uses ``self.time_properties`` and ``self.config`` to generate the hyperparameter grid.

        Converts model components and time properties into
        :class:`~greykite.sklearn.estimator.prophet_estimator.ProphetEstimator`
        hyperparameters.

        Returns
        -------
        hyperparameter_grid : `dict` [`str`, `list` [`any`]] or None
            :class:`~greykite.sklearn.estimator.prophet_estimator.ProphetEstimator`
            hyperparameters.

            See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.
            The output dictionary values are lists, combined in grid search.
        """
        self.config.model_components_param = self.apply_prophet_model_components_defaults(
            model_components=self.config.model_components_param,
            time_properties=self.time_properties)
        # Returns a set of parameters for grid search
        hyperparameter_grid = {
            "estimator__growth": self.config.model_components_param.growth["growth_term"],
            "estimator__seasonality_mode": self.config.model_components_param.seasonality["seasonality_mode"],
            "estimator__seasonality_prior_scale": self.config.model_components_param.seasonality["seasonality_prior_scale"],
            "estimator__yearly_seasonality": self.config.model_components_param.seasonality["yearly_seasonality"],
            "estimator__weekly_seasonality": self.config.model_components_param.seasonality["weekly_seasonality"],
            "estimator__daily_seasonality": self.config.model_components_param.seasonality["daily_seasonality"],
            "estimator__add_seasonality_dict": self.config.model_components_param.seasonality["add_seasonality_dict"],
            "estimator__holidays": [self.config.model_components_param.events["holidays_df"]],
            "estimator__holidays_prior_scale": self.config.model_components_param.events["holidays_prior_scale"],
            "estimator__changepoint_prior_scale": self.config.model_components_param.changepoints["changepoint_prior_scale"],
            "estimator__changepoints": self.config.model_components_param.changepoints["changepoints"],
            "estimator__n_changepoints": self.config.model_components_param.changepoints["n_changepoints"],
            "estimator__changepoint_range": self.config.model_components_param.changepoints["changepoint_range"],
            "estimator__mcmc_samples": self.config.model_components_param.uncertainty["mcmc_samples"],
            "estimator__uncertainty_samples": self.config.model_components_param.uncertainty["uncertainty_samples"],
            "estimator__add_regressor_dict": self.config.model_components_param.regressors["add_regressor_dict"]
        }

        # Overwrites values by `model_components.hyperparameter_override`
        # This may produce a list of dictionaries for grid search.
        hyperparameter_grid = update_dictionaries(
            hyperparameter_grid,
            overwrite_dicts=self.config.model_components_param.hyperparameter_override)

        # Ensures all items have the proper type for
        # `sklearn.model_selection.RandomizedSearchCV`.
        # List-type hyperparameters are specified below
        # with their accepted non-list type values.
        hyperparameter_grid = dictionaries_values_to_lists(
            hyperparameter_grid,
            hyperparameters_list_type={"estimator__changepoints": [None]}
        )
        return hyperparameter_grid

    def apply_template_decorator(func):
        """Decorator for ``apply_template_for_pipeline_params`` function.

        Overrides the method in `~greykite.framework.templates.base_template.BaseTemplate`.

        Raises
        ------
        ValueError if config.model_template != "PROPHET"
        """
        @functools.wraps(func)
        def process_wrapper(self, df: pd.DataFrame, config: Optional[ForecastConfig] = None):
            # sets defaults
            config = self.apply_forecast_config_defaults(config)
            # input validation
            if config.model_template != "PROPHET":
                if config.model_template != "PROPHET":
                    raise ValueError(f"ProphetTemplate only supports config.model_template='PROPHET', "
                                     f"found '{config.model_template}'")
            pipeline_params = func(self, df, config)
            return pipeline_params
        return process_wrapper

    @apply_template_decorator
    def apply_template_for_pipeline_params(
            self,
            df: pd.DataFrame,
            config: Optional[ForecastConfig] = None) -> Dict:
        """Explicitly calls the method in
        `~greykite.framework.templates.base_template.BaseTemplate`
        to make use of the decorator in this class.

        Parameters
        ----------
        df : `pandas.DataFrame`
            The time series dataframe with ``time_col`` and ``value_col`` and optional regressor columns.
        config : `~greykite.framework.templates.autogen.forecast_config.ForecastConfig`.
            The `ForecastConfig` class that includes model training parameters.

        Returns
        -------
        pipeline_parameters : `dict`
            The pipeline parameters consumable by
            `~greykite.framework.pipeline.pipeline.forecast_pipeline`.
        """
        return super().apply_template_for_pipeline_params(df=df, config=config)

    apply_template_decorator = staticmethod(apply_template_decorator)



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen
"""Main entry point to create a forecast.
Generates a forecast from input data and config and stores the result.
"""
import json
from enum import Enum
from typing import Dict
from typing import Optional
from typing import Type

import pandas as pd

from greykite.common.logging import LoggingLevelEnum
from greykite.common.logging import log_message
from greykite.common.python_utils import unique_elements_in_list
from greykite.framework.pipeline.pipeline import ForecastResult
from greykite.framework.pipeline.pipeline import forecast_pipeline
from greykite.framework.templates.autogen.forecast_config import ForecastConfig
from greykite.framework.templates.autogen.forecast_config import forecast_config_from_dict
from greykite.framework.templates.forecast_config_defaults import ForecastConfigDefaults
from greykite.framework.templates.model_templates import ModelTemplateEnum
from greykite.framework.templates.simple_silverkite_template import SimpleSilverkiteTemplate
from greykite.framework.templates.template_interface import TemplateInterface


class Forecaster:
    """The main entry point to creates a forecast.

    Call the :py:meth:`~greykite.framework.templates.forecaster.Forecaster.run_forecast_config`
    method to create a forecast. It takes a dataset and forecast configuration parameters.

    Notes
    -----
    This class can create forecasts using any of the model templates in
    `~greykite.framework.templates.model_templates.ModelTemplateEnum`.
    Model templates provide suitable default values for the available
    forecast estimators depending on the data characteristics.

    The model template is selected via the ``config.model_template``
    parameter to :py:meth:`~greykite.framework.templates.forecaster.Forecaster.run_forecast_config`.

    To add your own custom algorithms or template classes in our framework,
    pass ``model_template_enum`` and ``default_model_template_name``
    to the constructor.
    """
    def __init__(
            self,
            model_template_enum: Type[Enum] = ModelTemplateEnum,
            default_model_template_name: str = ModelTemplateEnum.SILVERKITE.name):
        # Optional user input
        self.model_template_enum: Type[Enum] = model_template_enum
        """The available template names. An Enum class where names are template names, and values are of type
        `~greykite.framework.templates.model_templates.ModelTemplate`.
        """
        self.default_model_template_name: str = default_model_template_name
        """The default template name if not provided by ``config.model_template``.
        Should be a name in ``model_template_enum``.
        Used by :py:meth:`~greykite.framework.templates.forecaster.Forecaster.__get_template_class`.
        """
        # The following are set by `self.run_forecast_config`.
        self.template_class: Optional[Type[TemplateInterface]] = None
        """Template class used. Must implement
        `~greykite.framework.templates.template_interface.TemplateInterface`
        and be one of the classes in ``self.model_template_enum``.
        Available for debugging purposes.
        Set by :py:meth:`~greykite.framework.templates.forecaster.Forecaster.run_forecast_config`.
        """
        self.template: Optional[TemplateInterface] = None
        """Instance of ``template_class`` used to run the forecast.
        See the docstring of the specific template class used.

            - `~greykite.framework.templates.simple_silverkite_template.SimpleSilverkiteTemplate`
            - `~greykite.framework.templates.silverkite_template.SilverkiteTemplate`
            - `~greykite.framework.templates.prophet_template.ProphetTemplate`
            - etc.

        Available for debugging purposes.
        Set by :py:meth:`~greykite.framework.templates.forecaster.Forecaster.run_forecast_config`.
        """
        self.config: Optional[ForecastConfig] = None
        """`~greykite.framework.templates.autogen.forecast_config.ForecastConfig`
        passed to the template class.
        Set by :py:meth:`~greykite.framework.templates.forecaster.Forecaster.run_forecast_config`.
        """
        self.pipeline_params: Optional[Dict] = None
        """Parameters used to call :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.
        Available for debugging purposes.
        Set by :py:meth:`~greykite.framework.templates.forecaster.Forecaster.run_forecast_config`.
        """
        self.forecast_result: Optional[ForecastResult] = None
        """The forecast result, returned by :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.
        Set by :py:meth:`~greykite.framework.templates.forecaster.Forecaster.run_forecast_config`.
        """

    def __get_config_with_default_model_template_and_components(self, config: Optional[ForecastConfig] = None) -> ForecastConfig:
        """Gets config with default value for `model_template` and `model_components_param` if not provided.

            - model_template : default value is ``self.default_model_template_name``.
            - model_components_param : default value is an empty ModelComponentsParam().

        Parameters
        ----------
        config : :class:`~greykite.framework.templates.model_templates.ForecastConfig` or None
            Config object for template class to use.
            See :class:`~greykite.framework.templates.model_templates.ForecastConfig`.
            If None, uses an empty ForecastConfig.

        Returns
        -------
        config : :class:`~greykite.framework.templates.model_templates.ForecastConfig`
            Input ``config`` with default ``model_template`` populated.
            If ``config.model_template`` is None, it is set to ``self.default_model_template_name``.
            If ``config.model_components_param`` is None, it is set to ``ModelComponentsParam()``.
        """
        config = config if config is not None else ForecastConfig()
        # Unpacks list of a single element and sets default value if None.
        # NB: Does not call `apply_forecast_config_defaults`.
        #   Only sets `model_template` and `model_components_param`.
        #   The template class may have its own implementation of forecast config defaults.
        forecast_config_defaults = ForecastConfigDefaults()
        forecast_config_defaults.DEFAULT_MODEL_TEMPLATE = self.default_model_template_name
        config.model_template = forecast_config_defaults.apply_model_template_defaults(config.model_template)
        config.model_components_param = forecast_config_defaults.apply_model_components_defaults(config.model_components_param)
        return config

    def __get_template_class(self, config: Optional[ForecastConfig] = None) -> Type[TemplateInterface]:
        """Extracts template class (e.g. `SimpleSilverkiteTemplate`) from the config.

        Parameters
        ----------
        config : :class:`~greykite.framework.templates.model_templates.ForecastConfig` or None
            Config object for template class to use.
            See :class:`~greykite.framework.templates.model_templates.ForecastConfig`.

        Returns
        -------
        template_class : Type[`~greykite.framework.templates.template_interface.TemplateInterface`]
            An implementation of `~greykite.framework.templates.template_interface.TemplateInterface`.
        """
        config = self.__get_config_with_default_model_template_and_components(config)

        if isinstance(config.model_template, list):
            # Parses `config.model_template` to extract the template class, with validation.
            # Handles a list of model templates.
            template_classes = [self.__get_template_class(config=ForecastConfig(model_template=mt))
                                for mt in config.model_template]
            for tc in template_classes:
                if tc != template_classes[0]:
                    raise ValueError("All model templates must use the same template class. "
                                     f"Found {template_classes}")
            template_class = template_classes[0]
            if not template_class().allow_model_template_list:
                raise ValueError(f"The template class {template_class} does not allow `model_template` to be a list. "
                                 f"Pass a string instead.")
        else:
            # Handles other situations (string, data class).
            try:
                # Tries to look up in `self.model_template_enum`.
                template_class = self.model_template_enum[config.model_template].value.template_class
            except (KeyError, TypeError):
                # Template is not found in the enum.
                # NB: The logic in this clause is written for the default `self.model_template_enum`,
                #   which contains only one template class that is a subclass of SimpleSilverkiteTemplate.
                #   If a custom `self.model_template_enum` is provided it may be useful to override this logic.
                valid_names = ", ".join(self.model_template_enum.__dict__["_member_names_"])
                # Checks if template enum has a template class that supports generic naming
                #   i.e. a subclass of `SimpleSilverkiteTemplate`.
                subclass_simple_silverkite = [mte for mte in self.model_template_enum
                                              if issubclass(mte.value.template_class, SimpleSilverkiteTemplate)]
                if len(subclass_simple_silverkite) > 0:
                    try:
                        log_message(f"Model template {config.model_template} is not found in the template enum. "
                                    f"Checking if model template is suitable for `SimpleSilverkiteTemplate`.", LoggingLevelEnum.DEBUG)
                        SimpleSilverkiteTemplate().check_template_type(config.model_template)
                        possible_template_classes = unique_elements_in_list([mte.value.template_class
                                                                             for mte in subclass_simple_silverkite])
                        if len(possible_template_classes) > 1:
                            log_message(f"Multiple template classes could be used for the model "
                                        f"template {config.model_template}: {possible_template_classes}", LoggingLevelEnum.DEBUG)
                        # arbitrarily take a class that supports generic naming
                        template_class = subclass_simple_silverkite[0].value.template_class
                        log_message(f"Using template class {template_class} for the model "
                                    f"template {config.model_template}", LoggingLevelEnum.DEBUG)
                    except ValueError:
                        raise ValueError(f"Model Template '{config.model_template}' is not recognized! Must be one of: {valid_names}"
                                         " or satisfy the `SimpleSilverkiteTemplate` rules.")
                else:
                    raise ValueError(f"Model Template '{config.model_template}' is not recognized! Must be one of: {valid_names}.")

        # Validates `model_components_param` compatibility with the template
        if not template_class().allow_model_components_param_list and isinstance(config.model_components_param, list):
            raise ValueError(f"Model template {config.model_template} does not support a list of `ModelComponentsParam`.")

        return template_class

    def apply_forecast_config(
            self,
            df: pd.DataFrame,
            config: Optional[ForecastConfig] = None) -> Dict:
        """Fetches pipeline parameters from the ``df`` and ``config``,
        but does not run the pipeline to generate a forecast.

        :py:meth:`~greykite.framework.templates.forecaster.Forecaster.run_forecast_config`
        calls this function and also runs the forecast pipeline.

        Available for debugging purposes to check pipeline parameters before
        running a forecast. Sets these attributes for debugging:

            - ``pipeline_params`` : the parameters passed to
              :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.
            - ``template_class``, ``template`` : the template class used to generate the
              pipeline parameters.
            - ``config`` : the :class:`~greykite.framework.templates.model_templates.ForecastConfig`
              passed as input to template class, to translate into pipeline parameters.

        Provides basic validation on the compatibility of ``config.model_template``
        with ``config.model_components_param``.

        Parameters
        ----------
        df : `pandas.DataFrame`
            Timeseries data to forecast.
            Contains columns [`time_col`, `value_col`], and optional regressor columns
            Regressor columns should include future values for prediction
        config : :class:`~greykite.framework.templates.model_templates.ForecastConfig` or None
            Config object for template class to use.
            See :class:`~greykite.framework.templates.model_templates.ForecastConfig`.

        Returns
        -------
        pipeline_params : `dict` [`str`, `any`]
            Input to :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.
        """
        self.config = self.__get_config_with_default_model_template_and_components(config)
        self.template_class = self.__get_template_class(self.config)
        self.template = self.template_class()
        self.pipeline_params = self.template.apply_template_for_pipeline_params(df=df, config=self.config)
        return self.pipeline_params

    def run_forecast_config(
            self,
            df: pd.DataFrame,
            config: Optional[ForecastConfig] = None) -> ForecastResult:
        """Creates a forecast from input data and config.
        The result is also stored as ``self.forecast_result``.

        Parameters
        ----------
        df : `pandas.DataFrame`
            Timeseries data to forecast.
            Contains columns [`time_col`, `value_col`], and optional regressor columns
            Regressor columns should include future values for prediction
        config : :class:`~greykite.framework.templates.model_templates.ForecastConfig`
            Config object for template class to use.
            See :class:`~greykite.framework.templates.model_templates.ForecastConfig`.

        Returns
        -------
        forecast_result : :class:`~greykite.framework.pipeline.pipeline.ForecastResult`
            Forecast result, an object of type
            :class:`~greykite.framework.pipeline.pipeline.ForecastResult`.

            The output of :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`,
            according to the ``df`` and ``config`` configuration parameters.
        """
        pipeline_parameters = self.apply_forecast_config(
            df=df,
            config=config)
        self.forecast_result = forecast_pipeline(**pipeline_parameters)
        return self.forecast_result

    def run_forecast_json(
            self,
            df: pd.DataFrame,
            json_str: str = "{}") -> ForecastResult:
        """Calls :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`
        according to the ``json_str`` configuration parameters.

        Parameters
        ----------
        df : `pandas.DataFrame`
            Timeseries data to forecast.
            Contains columns [`time_col`, `value_col`], and optional regressor columns
            Regressor columns should include future values for prediction
        json_str : `str`
            Json string of the config object for Forecast to use.
            See :class:`~greykite.framework.templates.model_templates.ForecastConfig`.

        Returns
        -------
        forecast_result : :class:`~greykite.framework.pipeline.pipeline.ForecastResult`
            Forecast result.
            The output of :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`,
            called using the template class with specified configuration.
            See :class:`~greykite.framework.pipeline.pipeline.ForecastResult`
            for details.
        """
        config_dict = json.loads(json_str)
        config = forecast_config_from_dict(config_dict)
        self.run_forecast_config(
            df=df,
            config=config)
        return self.forecast_result



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen
"""Enums for data frequency and seasonality."""

from collections import namedtuple
from enum import Enum


class TimeEnum(Enum):
    """Time constants"""
    ONE_WEEK_IN_DAYS = 7
    ONE_MONTH_IN_DAYS = 30
    ONE_QUARTER_IN_DAYS = 90
    ONE_YEAR_IN_DAYS = 365
    # Approximate number of seconds corresponding to each period.
    # May vary for leap year, daylight savings, etc.
    ONE_MINUTE_IN_SECONDS = 60
    ONE_HOUR_IN_SECONDS = 3600
    ONE_DAY_IN_SECONDS = 24 * 3600
    ONE_WEEK_IN_SECONDS = 7 * 24 * 3600
    ONE_MONTH_IN_SECONDS = 30 * 24 * 3600
    ONE_QUARTER_IN_SECONDS = 90 * 24 * 3600
    ONE_YEAR_IN_SECONDS = 365 * 24 * 3600


class SeasonalityEnum(Enum):
    """Valid types of seasonality available to use"""
    DAILY_SEASONALITY = "DAILY_SEASONALITY"
    WEEKLY_SEASONALITY = "WEEKLY_SEASONALITY"
    MONTHLY_SEASONALITY = "MONTHLY_SEASONALITY"
    QUARTERLY_SEASONALITY = "QUARTERLY_SEASONALITY"
    YEARLY_SEASONALITY = "YEARLY_SEASONALITY"


class FrequencyEnum(Enum):
    """Valid types of Frequency available to use"""
    MINUTE = "MINUTE"
    HOUR = "HOUR"
    DAY = "DAY"
    WEEK = "WEEK"
    MONTH = "MONTH"
    QUARTER = "QUARTER"
    YEAR = "YEAR"
    MULTIYEAR = "MULTIYEAR"


class SimpleTimeFrequencyEnum(Enum):
    """Provides default properties (horizon, seasonality) for various time frequencies.
    Used to define default values based on the closest frequency to the input data.
    """
    # default_horizon: how far ahead to forecast
    # seconds_per_observation: approximation of the period of each observation (in seconds)
    # valid_seas: valid seasonalities
    Frequency = namedtuple("Frequency", "default_horizon, seconds_per_observation, valid_seas")

    # minutely
    MINUTE = Frequency(
        default_horizon=60,
        seconds_per_observation=TimeEnum.ONE_MINUTE_IN_SECONDS.value,
        valid_seas={SeasonalityEnum.DAILY_SEASONALITY.name,
                    SeasonalityEnum.WEEKLY_SEASONALITY.name,
                    SeasonalityEnum.MONTHLY_SEASONALITY.name,
                    SeasonalityEnum.QUARTERLY_SEASONALITY.name,
                    SeasonalityEnum.YEARLY_SEASONALITY.name})

    # hourly
    HOUR = Frequency(
        default_horizon=24,
        seconds_per_observation=TimeEnum.ONE_HOUR_IN_SECONDS.value,
        valid_seas={SeasonalityEnum.DAILY_SEASONALITY.name,
                    SeasonalityEnum.WEEKLY_SEASONALITY.name,
                    SeasonalityEnum.MONTHLY_SEASONALITY.name,
                    SeasonalityEnum.QUARTERLY_SEASONALITY.name,
                    SeasonalityEnum.YEARLY_SEASONALITY.name})

    # daily
    # monthly seasonality is not used for hourly/daily data to avoid over-specification
    DAY = Frequency(
        default_horizon=30,
        seconds_per_observation=TimeEnum.ONE_DAY_IN_SECONDS.value,
        valid_seas={SeasonalityEnum.WEEKLY_SEASONALITY.name,
                    SeasonalityEnum.MONTHLY_SEASONALITY.name,
                    SeasonalityEnum.QUARTERLY_SEASONALITY.name,
                    SeasonalityEnum.YEARLY_SEASONALITY.name})

    # weekly
    WEEK = Frequency(
        default_horizon=12,
        seconds_per_observation=TimeEnum.ONE_WEEK_IN_SECONDS.value,
        valid_seas={SeasonalityEnum.MONTHLY_SEASONALITY.name,
                    SeasonalityEnum.QUARTERLY_SEASONALITY.name,
                    SeasonalityEnum.YEARLY_SEASONALITY.name})

    # monthly
    MONTH = Frequency(
        default_horizon=12,
        seconds_per_observation=TimeEnum.ONE_MONTH_IN_SECONDS.value,
        valid_seas={SeasonalityEnum.QUARTERLY_SEASONALITY.name,
                    SeasonalityEnum.YEARLY_SEASONALITY.name})

    # quarterly
    QUARTER = Frequency(
        default_horizon=12,
        seconds_per_observation=TimeEnum.ONE_QUARTER_IN_SECONDS.value,
        valid_seas={SeasonalityEnum.YEARLY_SEASONALITY.name})

    # yearly
    YEAR = Frequency(
        default_horizon=2,
        seconds_per_observation=TimeEnum.ONE_YEAR_IN_SECONDS.value,
        valid_seas={})

    # longer than yearly
    MULTIYEAR = Frequency(
        default_horizon=2,
        seconds_per_observation=TimeEnum.ONE_YEAR_IN_SECONDS.value * 2,  # not used anywhere, just a placeholder
        valid_seas={})



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Reza Hosseini
"""Calculates normal quantiles for each row of dataframe."""

import scipy


def normal_quantiles_df(
        df,
        std_col,
        mean_col=None,
        fixed_mean=0.0,
        quantiles=(0.025, 0.975)):
    """Calculates normal quantiles for each row of dataframe (df)
    using available means (either given in "mean_col" or "fixed_mean") and stds ("std_col").

    :param mean_col: string,  the column with the sample means for each row
    :param std: string, the column with the sample std for each row
    :param sample_size: integer, the sample size of the random sample for the given row
    :quantiles: tuple of float/double with values in [0, 1], denoting the desired quantiles
    """
    df = df.copy()
    if mean_col is None:
        df["normal_quantiles"] = df.apply(
            lambda row: scipy.stats.norm.ppf(
                loc=fixed_mean,
                scale=row[std_col],
                q=quantiles),
            axis=1)
    else:
        df["normal_quantiles"] = df.apply(
            lambda row: scipy.stats.norm.ppf(
                loc=row[mean_col],
                scale=row[std_col],
                q=quantiles),
            axis=1)

    return df



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Reza Hosseini
"""Functions that extract timeseries properties."""

import datetime
import warnings
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy as np
import pandas as pd

from greykite.common.constants import TIME_COL
from greykite.common.constants import VALUE_COL
from greykite.common.features.adjust_anomalous_data import adjust_anomalous_data
from greykite.common.logging import LoggingLevelEnum
from greykite.common.logging import log_message


def describe_timeseries(df, time_col):
    """Checks if a time series consists of equal time increments
    and if it is in increasing order.

    :param df: data.frame which includes the time column in datetime format
    :param time_col: time column

    :return: a dictionary with following items

        - dataframe ("df") with added delta columns
        - "regular_increments" booleans to see if time deltas as the same
        - "increasing" a boolean to denote if the time increments are increasing
        - "min_timestamp": minimum timestamp in data
        - "max_timestamp": maximum timestamp in data
        - "mean_increment_secs": mean of increments in seconds
        - "min_increment_secs": minimum of increments in seconds
        - "median_increment_secs": median of increments in seconds
        - "mean_delta": mean of time increments
        - "min_delta": min of the time increments
        - "max_delta": max of the time increments
        - "median_delta": median of the time increments
        - "freq_in_secs": the frequency of the timeseries in seconds which is defined to be the median time-gap
        - "freq_in_days": the frequency of the timeseries in days which is defined to be the median time-gap
        - "freq_in_timedelta": the frequency of the timeseries in `datetime.timedelta` which is defined to be the median time-gap

    """
    df = df.copy(deep=True)
    if df.shape[0] < 2:
        raise Exception("dataframe needs to have at least two rows")
    df["delta"] = (
            df[time_col] - df[time_col].shift()).fillna(pd.Timedelta(seconds=0))
    df["delta_sec"] = df["delta"].values / np.timedelta64(1, "s")

    delta_sec = df["delta_sec"][1:]
    mean_increment_secs = np.mean(delta_sec)
    min_increment_secs = min(delta_sec)
    median_increment_secs = np.median(delta_sec)
    regular_increments = (max(delta_sec) == min(delta_sec))
    increasing = min(delta_sec > 0)
    min_timestamp = df[time_col].min()
    max_timestamp = df[time_col].max()

    delta = df["delta"][1:]
    min_delta = min(delta)
    max_delta = max(delta)
    mean_delta = np.mean(delta)
    median_delta = np.median(delta)

    # The frequency is defined by the median time-gap
    freq_in_secs = median_increment_secs
    freq_in_days = freq_in_secs / (24 * 3600)
    freq_in_timedelta = datetime.timedelta(days=freq_in_days)

    return {
            "df": df,
            "regular_increments": regular_increments,
            "increasing": increasing,
            "min_timestamp": min_timestamp,
            "max_timestamp": max_timestamp,
            "mean_increment_secs": mean_increment_secs,
            "min_increment_secs": min_increment_secs,
            "median_increment_secs": median_increment_secs,
            "mean_delta": mean_delta,
            "median_delta": median_delta,
            "min_delta": min_delta,
            "max_delta": max_delta,
            "freq_in_secs": freq_in_secs,
            "freq_in_days": freq_in_days,
            "freq_in_timedelta": freq_in_timedelta
            }


def min_gap_in_seconds(df, time_col):
    """Returns the smallest gap between observations in df[time_col].

    Assumes df[time_col] is sorted in ascending order without duplicates.

    :param df: pd.DataFrame
        input timeseries
    :param time_col: str
        time column name in `df`
    :return: float
        minimum gap between observations, in seconds
    """
    if df.shape[0] < 2:
        raise ValueError(f"Must provide at least two data points. Found {df.shape[0]}.")
    timestamps = pd.to_datetime(df[time_col])
    period = (timestamps - timestamps.shift()).min()
    return period.days*24*3600 + period.seconds


def find_missing_dates(timestamp_series):
    """Identifies any gaps in pandas Series containing timestamps
    Timestamps are assumed to be in sorted order
    :return: pd.DataFrame with dates of the gaps in the time series, if any
    """
    timestamp_series = timestamp_series.sort_values(ascending=True, inplace=False)  # don't modify original
    delta = timestamp_series.diff()[1:].fillna(pd.Timedelta(seconds=0))
    delta_sec = delta.values / np.timedelta64(1, "s")
    min_gap = min(delta_sec)

    # dates surrounding the gap
    gap_start_points = timestamp_series[:-1][delta_sec > min_gap].reset_index(drop=True)
    gap_end_points = timestamp_series[1:][delta_sec > min_gap].reset_index(drop=True)
    # size of the gaps, in periods
    gap_periods = pd.Series((delta_sec[delta_sec > min_gap] / min_gap) - 1)
    gaps = pd.concat([gap_start_points, gap_end_points, gap_periods], axis=1, ignore_index=True)
    gaps.columns = ["right_before_gap", "right_after_gap", "gap_size"]
    return gaps


def fill_missing_dates(df, time_col=TIME_COL, freq=None):
    """Looks for gaps in df[time_col] and returns a pandas.DataFrame
        with the missing rows added in.
        Warning: if freq doesn't match intended freq, then values may be removed.

    Parameters
    ----------
    df : `pandas.DataFrame`
        dataframe with column ``time_col``
    time_col: `str`
        time column name, default TIME_COL
    freq: `str`
        timeseries frequency,
        DateOffset alias, default None (automatically inferred)

    Returns
    -------
    full_df : `pandas.DataFrame`
        ``df`` with rows added for missing timestamps
    added_timepoints : `int`
        The number of rows added to ``df``
    dropped_timepoints : `int`
        The number of rows removed from ``df``.
        If the timestamps in ``df`` are not evenly spaced,
        irregular timestamps may be removed.
    """
    freq = freq if freq is not None else pd.infer_freq(df[time_col])
    df = df.reset_index(drop=True)
    complete_dates = pd.DataFrame({
        time_col: pd.date_range(
            start=min(df[time_col]),
            end=max(df[time_col]),
            freq=freq)
    })
    full_df = pd.merge(complete_dates, df, how="left", on=time_col)

    # counts the timestamps in one but not the other
    before = set(df[time_col].values)
    after = set(full_df[time_col].values)
    added_timepoints = len(after - before)
    dropped_timepoints = len(before - after)
    if added_timepoints > 0:
        log_message(f"Added {added_timepoints} missing dates. There were {len(before)} values originally.",
                    LoggingLevelEnum.INFO)
    if dropped_timepoints > 0:
        warnings.warn(f"Dropped {dropped_timepoints} dates when filling gaps in input data. Provide data frequency"
                      f" and make sure data points are evenly spaced.")

    return full_df, added_timepoints, dropped_timepoints


def get_canonical_data(
        df: pd.DataFrame,
        time_col: str = TIME_COL,
        value_col: str = VALUE_COL,
        freq: str = None,
        date_format: str = None,
        tz: str = None,
        train_end_date: datetime = None,
        regressor_cols: List[str] = None,
        anomaly_info: Optional[Union[Dict, List[Dict]]] = None):
    """Loads data to internal representation. Parses date column,
    sets timezone aware index.
    Checks for irregularities and raises an error if input is invalid.
    Adjusts for anomalies according to ``anomaly_info``.

    Parameters
    ----------
    df : `pandas.DataFrame`
        Input timeseries. A data frame which includes the timestamp column
        as well as the value column.
    time_col : `str`
        The column name in ``df`` representing time for the time series data.
        The time column can be anything that can be parsed by pandas DatetimeIndex.
    value_col: `str`
        The column name which has the value of interest to be forecasted.
    freq : `str`, optional, default None
        Timeseries frequency, DateOffset alias, If None automatically inferred.
    date_format : `str`, optional, default None
        strftime format to parse time column, eg ``%m/%d/%Y``.
        Note that ``%f`` will parse all the way up to nanoseconds.
        If None (recommended), inferred by `pandas.to_datetime`.
    tz : `str` or pytz.timezone object, optional, default None
        Passed to `pandas.tz_localize` to localize the timestamp.
    train_end_date : `datetime.datetime`, optional, default None
        Last date to use for fitting the model. Forecasts are generated after this date.
        If None, it is set to the minimum of ``self.last_date_for_val`` and
        ``self.last_date_for_reg``.
    regressor_cols: `list` [`str`], optional, default None
        A list of regressor columns used in the training and prediction DataFrames.
        If None, no regressor columns are used.
        Regressor columns that are unavailable in ``df`` are dropped.
        anomaly_info : `dict` or None, default None
    anomaly_info : `dict` or `list` [`dict`] or None, default None
        Anomaly adjustment info. Anomalies in ``df``
        are corrected before any forecasting is done.

        If None, no adjustments are made.

        A dictionary containing the parameters to
        `~greykite.common.features.adjust_anomalous_data.adjust_anomalous_data`.
        See that function for details.
        The possible keys are:

            ``"value_col"`` : `str`
                The name of the column in ``df`` to adjust. You may adjust the value
                to forecast as well as any numeric regressors.
            ``"anomaly_df"`` : `pandas.DataFrame`
                Adjustments to correct the anomalies.
            ``"start_date_col"``: `str`, default START_DATE_COL
                Start date column in ``anomaly_df``.
            ``"end_date_col"``: `str`, default END_DATE_COL
                End date column in ``anomaly_df``.
            ``"adjustment_delta_col"``: `str` or None, default None
                Impact column in ``anomaly_df``.
            ``"filter_by_dict"``: `dict` or None, default None
                Used to filter ``anomaly_df`` to the relevant anomalies for
                the ``value_col`` in this dictionary.
                Key specifies the column name, value specifies the filter value.
            ``"filter_by_value_col""``: `str` or None, default None
                Adds ``{filter_by_value_col: value_col}`` to ``filter_by_dict``
                if not None, for the ``value_col`` in this dictionary.
            ``"adjustment_method"`` : `str` ("add" or "subtract"), default "add"
                How to make the adjustment, if ``adjustment_delta_col`` is provided.

        Accepts a list of such dictionaries to adjust multiple columns in ``df``.

    Returns
    -------
    canonical_data_dict : `dict`
        Dictionary containing the dataset in canonical form, and information such as
        train end date. Keys:

            ``"df"`` : `pandas.DataFrame`
                Data frame containing timestamp and value, with standardized column names for internal use
                (TIME_COL, VALUE_COL). Rows are sorted by time index, and missing gaps between dates are filled
                in so that dates are spaced at regular intervals. Values are adjusted for anomalies
                according to ``anomaly_info``.
                The index can be timezone aware (but TIME_COL is not).
            ``"df_before_adjustment"`` : `pandas.DataFrame` or None
                ``df`` before adjustment by ``anomaly_info``.
                If ``anomaly_info`` is None, this is None.
            ``"fit_df"`` : `pandas.DataFrame`
                A subset of the returned ``df``, with data up until ``train_end_date``.
            ``"freq"`` : `pandas.DataFrame`
                timeseries frequency, inferred if not provided
            ``"time_stats"`` : `dict`
                Information about the time column:

                    ``"gaps"``: missing_dates
                    ``"added_timepoints"``: added_timepoints
                    ``"dropped_timepoints"``: dropped_timepoints

            ``"regressor_cols"`` : `list` [`str`]
                A list of regressor columns.
            ``"fit_cols"`` : `list` [`str`]
                Names of time column, value column, and regressor columns.
            ``"train_end_date"`` : `datetime.datetime`
                Last date or timestamp for training. It is always less than or equal to
                minimum non-null values of ``last_date_for_val`` and ``last_date_for_reg``.
            ``"last_date_for_val"`` : `datetime.datetime`
                Date or timestamp corresponding  to last non-null value in ``df[value_col]``.
            ``"last_date_for_reg"`` : `datetime.datetime` or None
                Date or timestamp corresponding to last non-null value in ``df[regressor_cols]``.
                If ``regressor_cols`` is None, ``last_date_for_reg`` is None.
    """
    if time_col not in df.columns:
        raise ValueError(f"{time_col} column is not in input data")
    if value_col not in df.columns:
        raise ValueError(f"{value_col} column is not in input data")
    if df.shape[0] <= 2:
        raise ValueError(
            f"Time series has < 3 observations. More data are needed for forecasting.")

    # Standardizes the time column name.
    # `value_col` is standardized after anomalies are adjusted.
    df_standardized = df.rename({
        time_col: TIME_COL,
    }, axis=1)
    df_standardized[TIME_COL] = pd.to_datetime(
        df_standardized[TIME_COL],
        format=date_format,
        infer_datetime_format=True)
    # Drops data points from duplicate time stamps
    df_standardized.drop_duplicates(
        subset=[TIME_COL],
        keep='first',
        inplace=True)
    if df.shape[0] > df_standardized.shape[0]:
        warnings.warn(
            f"Duplicate timestamps have been removed.",
            UserWarning)
    df = df_standardized.sort_values(by=TIME_COL)
    # Infers data frequency
    inferred_freq = pd.infer_freq(df[TIME_COL])
    if freq is None:
        freq = inferred_freq
    elif inferred_freq is not None and freq != inferred_freq:
        warnings.warn(
            f"Provided frequency '{freq}' does not match inferred frequency '{inferred_freq}'."
            f" Using '{freq}'.", UserWarning)  # NB: with missing data, it's better to provide freq
    # Handles gaps in time series
    missing_dates = find_missing_dates(df[TIME_COL])
    df, added_timepoints, dropped_timepoints = fill_missing_dates(
        df,
        time_col=TIME_COL,
        freq=freq)
    time_stats = {
        "gaps": missing_dates,
        "added_timepoints": added_timepoints,
        "dropped_timepoints": dropped_timepoints
    }
    # Creates index with localized timestamp
    df.index = df[TIME_COL]
    df.index.name = None
    if tz is not None:
        df = df.tz_localize(tz)

    df_before_adjustment = None
    if anomaly_info is not None:
        # Saves values before adjustment.
        df_before_adjustment = df.copy()
        # Adjusts columns in df (e.g. `value_col`, `regressor_cols`)
        # using the anomaly info. One dictionary of parameters
        # for `adjust_anomalous_data` is provided for each column to adjust.
        if not isinstance(anomaly_info, (list, tuple)):
            anomaly_info = [anomaly_info]
        for single_anomaly_info in anomaly_info:
            adjusted_df_dict = adjust_anomalous_data(
                df=df,
                time_col=TIME_COL,
                **single_anomaly_info)
            # `self.df` with values for single_anomaly_info["value_col"] adjusted.
            df = adjusted_df_dict["adjusted_df"]

        # Standardizes `value_col` name
        df_before_adjustment.rename({
            value_col: VALUE_COL
        }, axis=1, inplace=True)
    # Standardizes `value_col` name
    df.rename({
        value_col: VALUE_COL
    }, axis=1, inplace=True)

    # Finds date of last available value
    last_date_available = df[TIME_COL].max()
    last_date_for_val = df[df[VALUE_COL].notnull()][TIME_COL].max()
    last_date_for_reg = None
    if regressor_cols:
        available_regressor_cols = [col for col in df.columns if col not in [TIME_COL, VALUE_COL]]
        cols_not_selected = set(regressor_cols) - set(available_regressor_cols)
        regressor_cols = [col for col in regressor_cols if col in available_regressor_cols]
        if cols_not_selected:
            warnings.warn(f"The following columns are not available to use as "
                          f"regressors: {sorted(cols_not_selected)}")
        last_date_for_reg = df[df[regressor_cols].notnull().any(axis=1)][TIME_COL].max()
        max_train_end_date = min(last_date_for_val, last_date_for_reg)
    else:
        max_train_end_date = last_date_for_val
    # Chooses appropriate train_end_date
    if train_end_date is None:
        train_end_date = max_train_end_date
        if train_end_date < last_date_available:
            warnings.warn(
                f"{value_col} column of the provided TimeSeries contains "
                f"null values at the end. Setting 'train_end_date' to the last timestamp with a "
                f"non-null value ({train_end_date}).",
                UserWarning)
    elif train_end_date > max_train_end_date:
        warnings.warn(
            f"Input timestamp for the parameter 'train_end_date' "
            f"({train_end_date}) either exceeds the last available timestamp or"
            f"{value_col} column of the provided TimeSeries contains null "
            f"values at the end. Setting 'train_end_date' to the last timestamp with a "
            f"non-null value ({max_train_end_date}).",
            UserWarning)
        train_end_date = max_train_end_date

    if regressor_cols is None:
        regressor_cols = []
    fit_cols = [TIME_COL, VALUE_COL] + regressor_cols
    fit_df = df[df[TIME_COL] <= train_end_date][fit_cols]

    return {
        "df": df,
        "df_before_adjustment": df_before_adjustment,
        "fit_df": fit_df,
        "freq": freq,
        "time_stats": time_stats,
        "regressor_cols": regressor_cols,
        "fit_cols": fit_cols,
        "train_end_date": train_end_date,
        "last_date_for_val": last_date_for_val,
        "last_date_for_reg": last_date_for_reg,
    }



# flake8: noqa
# To use this code, make sure you
#
#     import json
#
# and then, to convert JSON from a string, do
#
#     result = forecast_config_from_dict(json.loads(json_string))

from dataclasses import dataclass
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import TypeVar
from typing import Union
from typing import cast


T = TypeVar("T")


def from_int(x: Any) -> int:
    assert isinstance(x, int) and not isinstance(x, bool)
    return x


def from_none(x: Any) -> Any:
    assert x is None
    return x


def from_union(fs, x):
    for f in fs:
        try:
            return f(x)
        except:
            pass
    assert False


def from_dict(f: Callable[[Any], T], x: Any) -> Dict[str, T]:
    assert isinstance(x, dict)
    return { k: f(v) for (k, v) in x.items() }


def from_list_dict(f: Callable[[Any], T], x: Any) -> List[Dict[str, T]]:
    """Parses list of dictionaries, applying `f` to the dictionary values.
    All items must be dictionaries.
    """
    assert isinstance(x, list)
    assert all(isinstance(d, dict) for d in x)
    return [ { k: f(v) for (k, v) in d.items() } for d in x]


def from_list_dict_or_none(f: Callable[[Any], T], x: Any) -> List[Optional[Dict[str, T]]]:
    """Parses list of dictionaries or None elements, applying `f` to the dictionary values.
    If an element in the list is None, it is returned directly.
    """
    assert isinstance(x, list)
    assert all(d is None or isinstance(d, dict)for d in x)
    return [ { k: f(v) for (k, v) in d.items() } if isinstance(d, dict) else d for d in x]


def from_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def from_list_str(x: Any) -> List[str]:
    assert isinstance(x, list)
    assert all(isinstance(item, str) for item in x)
    return x


def from_float(x: Any) -> float:
    assert isinstance(x, (float, int)) and not isinstance(x, bool)
    return float(x)


def to_float(x: Any) -> float:
    assert isinstance(x, float)
    return x


def from_bool(x: Any) -> bool:
    assert isinstance(x, bool)
    return x


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


@dataclass
class ComputationParam:
    """How to compute the result."""
    hyperparameter_budget: Optional[int] = None
    """See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`."""
    n_jobs: Optional[int] = None
    """See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`."""
    verbose: Optional[int] = None
    """See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`."""

    @staticmethod
    def from_dict(obj: Any) -> 'ComputationParam':
        assert isinstance(obj, dict)
        hyperparameter_budget = from_union([from_int, from_none], obj.get("hyperparameter_budget"))
        n_jobs = from_union([from_int, from_none], obj.get("n_jobs"))
        verbose = from_union([from_int, from_none], obj.get("verbose"))
        return ComputationParam(
            hyperparameter_budget=hyperparameter_budget,
            n_jobs=n_jobs,
            verbose=verbose)

    def to_dict(self) -> dict:
        result: dict = {}
        result["hyperparameter_budget"] = from_union([from_int, from_none], self.hyperparameter_budget)
        result["n_jobs"] = from_union([from_int, from_none], self.n_jobs)
        result["verbose"] = from_union([from_int, from_none], self.verbose)
        return result


@dataclass
class EvaluationMetricParam:
    """What metrics to evaluate"""
    agg_func: Optional[Callable] = None
    """See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`."""
    agg_periods: Optional[int] = None
    """See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`."""
    cv_report_metrics: Optional[Union[str, List[str]]] = None
    """See `score_func` in :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`."""
    cv_selection_metric: Optional[str] = None
    """See `score_func` in :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`."""
    null_model_params: Optional[Dict[str, Any]] = None
    """See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`."""
    relative_error_tolerance: Optional[float] = None
    """See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`."""

    @staticmethod
    def from_dict(obj: Any) -> 'EvaluationMetricParam':
        assert isinstance(obj, dict)
        agg_func = from_union([lambda x: from_dict(lambda x: x, x), from_none], obj.get("agg_func"))
        agg_periods = from_union([from_int, from_none], obj.get("agg_periods"))
        cv_report_metrics = from_union([from_str, from_list_str, from_none], obj.get("cv_report_metrics"))
        cv_selection_metric = from_union([from_str, from_none], obj.get("cv_selection_metric"))
        null_model_params = from_union([lambda x: from_dict(lambda x: x, x), from_none], obj.get("null_model_params"))
        relative_error_tolerance = from_union([from_float, from_none], obj.get("relative_error_tolerance"))
        return EvaluationMetricParam(
            agg_func=agg_func,
            agg_periods=agg_periods,
            cv_report_metrics=cv_report_metrics,
            cv_selection_metric=cv_selection_metric,
            null_model_params=null_model_params,
            relative_error_tolerance=relative_error_tolerance)

    def to_dict(self) -> dict:
        result: dict = {}
        result["agg_func"] = from_union([lambda x: from_dict(lambda x: x, x), from_none], self.agg_func)
        result["agg_periods"] = from_union([from_int, from_none], self.agg_periods)
        result["cv_report_metrics"] = from_union([from_str, from_list_str, from_none], self.cv_report_metrics)
        result["cv_selection_metric"] = from_union([from_str, from_none], self.cv_selection_metric)
        result["null_model_params"] = from_union([lambda x: from_dict(lambda x: x, x), from_none], self.null_model_params)
        result["relative_error_tolerance"] = from_union([to_float, from_none], self.relative_error_tolerance)
        return result


@dataclass
class EvaluationPeriodParam:
    """How to split data for evaluation."""
    cv_expanding_window: Optional[bool] = None
    """See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`."""
    cv_horizon: Optional[int] = None
    """See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`."""
    cv_max_splits: Optional[int] = None
    """See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`."""
    cv_min_train_periods: Optional[int] = None
    """See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`."""
    cv_periods_between_splits: Optional[int] = None
    """See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`."""
    cv_periods_between_train_test: Optional[int] = None
    """See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`."""
    cv_use_most_recent_splits: Optional[bool] = None
    """See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`."""
    periods_between_train_test: Optional[int] = None
    """See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`"""
    test_horizon: Optional[int] = None
    """See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`."""

    @staticmethod
    def from_dict(obj: Any) -> 'EvaluationPeriodParam':
        assert isinstance(obj, dict)
        cv_expanding_window = from_union([from_bool, from_none], obj.get("cv_expanding_window"))
        cv_horizon = from_union([from_int, from_none], obj.get("cv_horizon"))
        cv_max_splits = from_union([from_int, from_none], obj.get("cv_max_splits"))
        cv_min_train_periods = from_union([from_int, from_none], obj.get("cv_min_train_periods"))
        cv_periods_between_splits = from_union([from_int, from_none], obj.get("cv_periods_between_splits"))
        cv_periods_between_train_test = from_union([from_int, from_none], obj.get("cv_periods_between_train_test"))
        cv_use_most_recent_splits = from_union([from_bool, from_none], obj.get("cv_use_most_recent_splits"))
        periods_between_train_test = from_union([from_int, from_none], obj.get("periods_between_train_test"))
        test_horizon = from_union([from_int, from_none], obj.get("test_horizon"))
        return EvaluationPeriodParam(
            cv_expanding_window=cv_expanding_window,
            cv_horizon=cv_horizon,
            cv_max_splits=cv_max_splits,
            cv_min_train_periods=cv_min_train_periods,
            cv_periods_between_splits=cv_periods_between_splits,
            cv_periods_between_train_test=cv_periods_between_train_test,
            cv_use_most_recent_splits=cv_use_most_recent_splits,
            periods_between_train_test=periods_between_train_test,
            test_horizon=test_horizon)

    def to_dict(self) -> dict:
        result: dict = {}
        result["cv_expanding_window"] = from_union([from_bool, from_none], self.cv_expanding_window)
        result["cv_horizon"] = from_union([from_int, from_none], self.cv_horizon)
        result["cv_max_splits"] = from_union([from_int, from_none], self.cv_max_splits)
        result["cv_min_train_periods"] = from_union([from_int, from_none], self.cv_min_train_periods)
        result["cv_periods_between_splits"] = from_union([from_int, from_none], self.cv_periods_between_splits)
        result["cv_periods_between_train_test"] = from_union([from_int, from_none], self.cv_periods_between_train_test)
        result["cv_use_most_recent_splits"] = from_union([from_bool, from_none], self.cv_use_most_recent_splits)
        result["periods_between_train_test"] = from_union([from_int, from_none], self.periods_between_train_test)
        result["test_horizon"] = from_union([from_int, from_none], self.test_horizon)
        return result


@dataclass
class MetadataParam:
    """Properties of the input data"""
    anomaly_info: Optional[Union[Dict[str, Any], List[Dict[str, Any]]]] = None
    """Anomaly adjustment info. Anomalies in ``df`` are corrected before any forecasting is
    done. If None, no adjustments are made.
    See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.
    """
    date_format: Optional[str] = None
    """See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`"""
    freq: Optional[str] = None
    """See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`"""
    time_col: Optional[str] = None
    """See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`"""
    train_end_date: Optional[str] = None
    """Last date to use for fitting the model. Forecasts are generated after this date.
    If None, it is set to the last date with a non-null value in value_col df.
    See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.
    """
    value_col: Optional[str] = None
    """See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`"""

    @staticmethod
    def from_dict(obj: Any) -> 'MetadataParam':
        assert isinstance(obj, dict)
        anomaly_info = from_union([
            lambda x: from_dict(lambda x: x, x),
            lambda x: from_list_dict(lambda x: x, x),
            from_none], obj.get("anomaly_info"))
        date_format = from_union([from_str, from_none], obj.get("date_format"))
        freq = from_union([from_str, from_none], obj.get("freq"))
        time_col = from_union([from_str, from_none], obj.get("time_col"))
        train_end_date = from_union([from_str, from_none], obj.get("train_end_date"))
        value_col = from_union([from_str, from_none], obj.get("value_col"))
        return MetadataParam(anomaly_info, date_format, freq, time_col, train_end_date, value_col)

    def to_dict(self) -> dict:
        result: dict = {}
        result["anomaly_info"] = from_union([
            lambda x: from_dict(lambda x: x, x),
            lambda x: from_list_dict(lambda x: x, x),
            from_none], self.anomaly_info)
        result["date_format"] = from_union([from_str, from_none], self.date_format)
        result["freq"] = from_union([from_str, from_none], self.freq)
        result["time_col"] = from_union([from_str, from_none], self.time_col)
        result["train_end_date"] = from_union([from_str, from_none], self.train_end_date)
        result["value_col"] = from_union([from_str, from_none], self.value_col)
        return result


@dataclass
class ModelComponentsParam:
    """Parameters to tune the model."""
    autoregression: Optional[Dict[str, Any]] = None
    """For modeling autoregression, see template for details"""
    changepoints: Optional[Dict[str, Any]] = None
    """For modeling changepoints, see template for details"""
    custom: Optional[Dict[str, Any]] = None
    """Additional parameters used by template, see template for details"""
    events: Optional[Dict[str, Any]] = None
    """For modeling events, see template for details"""
    growth: Optional[Dict[str, Any]] = None
    """For modeling growth (trend), see template for details"""
    hyperparameter_override: Optional[Union[Dict, List[Optional[Dict]]]] = None
    """After the above model components are used to create a hyperparameter grid,
    the result is updated by this dictionary, to create new keys or override existing ones.
    Allows for complete customization of the grid search.
    """
    regressors: Optional[Dict[str, Any]] = None
    """For modeling regressors, see template for details"""
    seasonality: Optional[Dict[str, Any]] = None
    """For modeling seasonality, see template for details"""
    uncertainty: Optional[Dict[str, Any]] = None
    """For modeling uncertainty, see template for details"""

    @staticmethod
    def from_dict(obj: Any) -> 'ModelComponentsParam':
        assert isinstance(obj, dict)
        autoregression = from_union([lambda x: from_dict(lambda x: x, x), from_none], obj.get("autoregression"))
        changepoints = from_union([lambda x: from_dict(lambda x: x, x), from_none], obj.get("changepoints"))
        custom = from_union([lambda x: from_dict(lambda x: x, x), from_none], obj.get("custom"))
        events = from_union([lambda x: from_dict(lambda x: x, x), from_none], obj.get("events"))
        growth = from_union([lambda x: from_dict(lambda x: x, x), from_none], obj.get("growth"))
        hyperparameter_override = from_union([
            lambda x: from_dict(lambda x: x, x),
            lambda x: from_list_dict_or_none(lambda x: x, x),
            from_none], obj.get("hyperparameter_override"))
        regressors = from_union([lambda x: from_dict(lambda x: x, x), from_none], obj.get("regressors"))
        seasonality = from_union([lambda x: from_dict(lambda x: x, x), from_none], obj.get("seasonality"))
        uncertainty = from_union([lambda x: from_dict(lambda x: x, x), from_none], obj.get("uncertainty"))
        return ModelComponentsParam(
            autoregression=autoregression,
            changepoints=changepoints,
            custom=custom,
            events=events,
            growth=growth,
            hyperparameter_override=hyperparameter_override,
            regressors=regressors,
            seasonality=seasonality,
            uncertainty=uncertainty)

    def to_dict(self) -> dict:
        result: dict = {}
        result["autoregression"] = from_union([lambda x: from_dict(lambda x: x, x), from_none], self.autoregression)
        result["changepoints"] = from_union([lambda x: from_dict(lambda x: x, x), from_none], self.changepoints)
        result["custom"] = from_union([lambda x: from_dict(lambda x: x, x), from_none], self.custom)
        result["events"] = from_union([lambda x: from_dict(lambda x: x, x), from_none], self.events)
        result["growth"] = from_union([lambda x: from_dict(lambda x: x, x), from_none], self.growth)
        result["hyperparameter_override"] = from_union([
            lambda x: from_dict(lambda x: x, x),
            lambda x: from_list_dict_or_none(lambda x: x, x),
            from_none], self.hyperparameter_override)
        result["regressors"] = from_union([lambda x: from_dict(lambda x: x, x), from_none], self.regressors)
        result["seasonality"] = from_union([lambda x: from_dict(lambda x: x, x), from_none], self.seasonality)
        result["uncertainty"] = from_union([lambda x: from_dict(lambda x: x, x), from_none], self.uncertainty)
        return result


@dataclass
class ForecastConfig:
    """Config for providing parameters to the Forecast library"""
    computation_param: Optional[ComputationParam] = None
    """How to compute the result. See
    :class:`~greykite.framework.templates.autogen.forecast_config.ComputationParam`.
    """
    coverage: Optional[float] = None
    """Intended coverage of the prediction bands (0.0 to 1.0).
    If None, the upper/lower predictions are not returned.
    """
    evaluation_metric_param: Optional[EvaluationMetricParam] = None
    """What metrics to evaluate. See
    :class:`~greykite.framework.templates.autogen.forecast_config.EvaluationMetricParam`.
    """
    evaluation_period_param: Optional[EvaluationPeriodParam] = None
    """How to split data for evaluation. See
    :class:`~greykite.framework.templates.autogen.forecast_config.EvaluationPeriodParam`.
    """
    forecast_horizon: Optional[int] = None
    """Number of periods to forecast into the future. Must be > 0.
    If None, default is determined from input data frequency.
    """
    metadata_param: Optional[MetadataParam] = None
    """Information about the input data. See
    :class:`~greykite.framework.templates.autogen.forecast_config.MetadataParam`.
    """
    model_components_param: Optional[Union[ModelComponentsParam, List[Optional[ModelComponentsParam]]]] = None
    """Parameters to tune the model. Typically a single ModelComponentsParam, but the `SimpleSilverkiteTemplate`
    template also allows a list of ModelComponentsParam for grid search. A single ModelComponentsParam
    corresponds to one grid, and a list corresponds to a list of grids.
    See :class:`~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam`.
    """
    model_template: Optional[Union[str, dataclass, List[Union[str, dataclass]]]] = None
    """Name of the model template. Typically a single string, but the `SimpleSilverkiteTemplate`
    template also allows a list of string for grid search.
    See :class:`~greykite.framework.templates.model_templates.ModelTemplateEnum`
    for valid names.
    """

    @staticmethod
    def from_dict(obj: Any) -> 'ForecastConfig':
        assert isinstance(obj, dict)
        computation_param = from_union([ComputationParam.from_dict, from_none], obj.get("computation_param"))
        coverage = from_union([from_float, from_none], obj.get("coverage"))
        evaluation_metric_param = from_union([EvaluationMetricParam.from_dict, from_none], obj.get("evaluation_metric_param"))
        evaluation_period_param = from_union([EvaluationPeriodParam.from_dict, from_none], obj.get("evaluation_period_param"))
        forecast_horizon = from_union([from_int, from_none], obj.get("forecast_horizon"))
        metadata_param = from_union([MetadataParam.from_dict, from_none], obj.get("metadata_param"))
        if not isinstance(obj.get("model_components_param"), list):
            obj["model_components_param"] = [obj.get("model_components_param")]
        model_components_param = [from_union([ModelComponentsParam.from_dict, from_none], mcp) for mcp in obj.get("model_components_param")]
        if not isinstance(obj.get("model_template"), list):
            obj["model_template"] = [obj.get("model_template")]
        model_template = [from_union([from_str, from_none], mt) for mt in obj.get("model_template")]
        return ForecastConfig(
            computation_param=computation_param,
            coverage=coverage,
            evaluation_metric_param=evaluation_metric_param,
            evaluation_period_param=evaluation_period_param,
            forecast_horizon=forecast_horizon,
            metadata_param=metadata_param,
            model_components_param=model_components_param,
            model_template=model_template)

    def to_dict(self) -> dict:
        result: dict = {}
        result["computation_param"] = from_union([lambda x: to_class(ComputationParam, x), from_none], self.computation_param)
        result["coverage"] = from_union([to_float, from_none], self.coverage)
        result["evaluation_metric_param"] = from_union([lambda x: to_class(EvaluationMetricParam, x), from_none], self.evaluation_metric_param)
        result["evaluation_period_param"] = from_union([lambda x: to_class(EvaluationPeriodParam, x), from_none], self.evaluation_period_param)
        result["forecast_horizon"] = from_union([from_int, from_none], self.forecast_horizon)
        result["metadata_param"] = from_union([lambda x: to_class(MetadataParam, x), from_none], self.metadata_param)
        if not isinstance(self.model_components_param, list):
            self.model_components_param = [self.model_components_param]
        result["model_components_param"] = [from_union([lambda x: to_class(ModelComponentsParam, x), from_none], mcp) for mcp in self.model_components_param]
        if not isinstance(self.model_template, list):
            self.model_template = [self.model_template]
        result["model_template"] = [from_union([from_str, from_none], mt) for mt in self.model_template]
        return result


def forecast_config_from_dict(s: Any) -> ForecastConfig:
    return ForecastConfig.from_dict(s)


def forecast_config_to_dict(x: ForecastConfig) -> Any:
    return to_class(ForecastConfig, x)



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Reza Hosseini
"""Pandas dataframe utils for uncertainty interval calculation."""


def offset_tuple_col(df, offset_col, tuple_col):
    """Takes a dataframe (df) with at least two columns
    (`offset_col` and `tuple_col`)
    and offsets all values given in tuple_col using an offset

    Parameters
    ----------
    df: `pd.DataFrame`
        input dataframe which includes desired values
    offset_col: `str`
        The column with the offset value
    tuple_col: `str`
        The column with tuples including real (float/double/int) numbers

    Returns
    -------
    df_offset : `pd.Dataframe`
        A dataframe which is same as ``df`` and only the
        ``tuple_col`` is altered by adding an offset to
        all values in each tuple.
    """
    return df.apply(
        lambda row: tuple(
            e + row[offset_col] for e in row[tuple_col]),
        axis=1)


def limit_tuple_col(df, tuple_col, lower=None, upper=None):
    """This function applies limits (lower bound and upper bound)
    to values (could be any value type for which min and max are defined)
    given in tuples appearing in `tuple_col` column of dataframe `df`.

    Parameters
    ----------
    df: `pd.DataFrame`
        Input DataFrame
    tuple_col: `str`
        The column of interest consisting of tuples
    lower: `int` or `float` or `double` or `str`
        Any type for which order "<" is defined.
        This input has the same type of values in the tuples,
        and denotes the desired lower bound
    upper: `int` or `float` or `double` or `str`
        Any type for which order "<" is defined.
        This input has the same type of values in the tuples,
        and denotes the desired upper bound

    Returns
    -------
    df : `pd.Dataframe`
        Dataframe of same shape as `df` with only values in `tuple_col` altered
        to satisfy the bounds: `lower`, `upper`
    """
    if lower is not None:
        df[tuple_col] = df.apply(
            lambda row: tuple(
                max(e, lower) for e in row[tuple_col]),
            axis=1)

    if upper is not None:
        df[tuple_col] = df.apply(
            lambda row: tuple(
                min(e, upper) for e in row[tuple_col]),
            axis=1)
    return df
